/* gjiten_settings.c
   
   GJITEN : A GTK+/GNOME BASED JAPANESE DICTIONARY
  
   Copyright (C) 1999-2003 Botond Botyanszki <boti at rocketmail dot com>
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published  by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <libgnome/libgnome.h>

#define __GJITEN_SETTINGS__
#include "gjiten-settings.h"
#include "gjiten.h"
#include "config.h"

extern gchar *kanjidicstrg[];
extern gchar *gnome_dialog_msg;

int load_settings() {
  gchar dicprefix[] = "/apps/gjiten/general/dic";
  gchar dicstr[25];
  gchar historystr[30];
  gchar *tmpptr;
  gchar gnomekcfg[] = "/apps/gjiten/kanjidic/";
  int i;

  conf.version = gconf_client_get_string(gconf_client, "/apps/gjiten/general/version", NULL);
  if (conf.version == NULL) { // First run
    conf.kdiccfg[KANJI] = TRUE; 
    conf.kdiccfg[RADICAL] = TRUE;
    conf.kdiccfg[STROKES] = TRUE;
    conf.kdiccfg[READING] = TRUE;
    conf.kdiccfg[ENGLISH] = TRUE;
    conf.kdiccfg[FREQ] = TRUE;   
    conf.kdiccfg[JOUYOU] = TRUE;
    conf.kdiccfg[CREF] = TRUE;
    conf.usekanjipad = FALSE;
    conf.usekanjidic = TRUE;
    conf.toolbar = TRUE;
    conf.menubar = TRUE;
    //conf.tooltips = TRUE;
    conf.force_ja_JP = TRUE;
    conf.force_language_c = TRUE;
    conf.kanjipad = PREFIX"/bin/kanjipad";
    conf.kanjidicfile = DEFDICDIR"kanjidic"; 
    conf.dictpath = DEFDICDIR;
    conf.searchlimit_enabled = FALSE;
    conf.maxwordmatches = DEFMAXWORDMATCHES;
    conf.autoadjust_enabled = TRUE;
    return 1; // no cfg
  }

  conf.autoadjust_enabled = gconf_client_get_bool(gconf_client, "/apps/gjiten/general/autoadjust_enabled", NULL);
  conf.deinflection_enabled = gconf_client_get_bool(gconf_client, "/apps/gjiten/general/deinflection_enabled", NULL);

  conf.bigwords = gconf_client_get_bool(gconf_client, "/apps/gjiten/general/bigwords", NULL);
  conf.bigkanji = gconf_client_get_bool(gconf_client, "/apps/gjiten/general/bigkanji", NULL);
  conf.bigfont = gconf_client_get_string(gconf_client, "/apps/gjiten/general/bigfont", NULL);
  conf.normalfont = gconf_client_get_string(gconf_client, "/apps/gjiten/general/normalfont", NULL);
  conf.gdk_use_xft = gconf_client_get_bool(gconf_client, "/apps/gjiten/general/gdk_use_xft", NULL);
  conf.force_ja_JP = gconf_client_get_bool(gconf_client, "/apps/gjiten/general/force_ja_JP", NULL);
  conf.force_language_c = gconf_client_get_bool(gconf_client, "/apps/gjiten/general/force_language_c", NULL);

  conf.searchlimit_enabled = gconf_client_get_bool(gconf_client, "/apps/gjiten/general/searchlimit_enabled", NULL);
  conf.maxwordmatches = gconf_client_get_int(gconf_client, "/apps/gjiten/general/maxwordmatches", NULL);
  conf.dictpath = gconf_client_get_string(gconf_client, "/apps/gjiten/general/dictpath", NULL);
  //conf.tooltips = gconf_client_get_bool(gconf_client, "/apps/gjiten/general/tooltips", NULL);
  conf.menubar = gconf_client_get_bool(gconf_client, "/apps/gjiten/general/menubar", NULL);
  conf.toolbar = gconf_client_get_bool(gconf_client, "/apps/gjiten/general/toolbar", NULL);
  conf.usekanjipad = gconf_client_get_bool(gconf_client, "/apps/gjiten/general/usekanjipad", NULL);
  conf.usekanjidic = gconf_client_get_bool(gconf_client, "/apps/gjiten/general/usekanjidic", NULL);
  conf.kanjidicfile = gconf_client_get_string(gconf_client, "/apps/gjiten/kanjidic/kanjidicfile", NULL);
  if (conf.kanjidicfile == NULL) {
    conf.usekanjidic = FALSE; 
    conf.kanjidicfile = DEFDICDIR"kanjidic";
    //printf("NULL %s\n", DEFDICDIR"kanjidic");
  }
  else if (strlen(conf.kanjidicfile) == 0) {
    conf.usekanjidic = FALSE;
    conf.kanjidicfile = DEFDICDIR"kanjidic";
    //printf("0 %s\n", DEFDICDIR"kanjidic");
  }
  conf.kanjipad = gconf_client_get_string(gconf_client, "/apps/gjiten/general/kanjipad", NULL);
  conf.numofdics = gconf_client_get_int(gconf_client, "/apps/gjiten/general/numofdics", NULL);
  for (i = 0; i < conf.numofdics; i++) {
    if (i == MAXDICFILES - 1) break;
    sprintf(dicstr, "%s%d", dicprefix, i);
    conf.dicfiles[i] = gconf_client_get_string(gconf_client, dicstr, NULL);
    if (conf.dicfiles[i] == NULL) {
      conf.numofdics = i;
      g_error("Dicfile numbers and dicfile strings inconsistent in GConf");
    }
  }

  //Load kanjidic display settings
  for (i = 0; i < KCFGNUM; i++) { 
    tmpptr = g_malloc(strlen(gnomekcfg) + strlen(kanjidicstrg[i]) + 1);
    strcpy(tmpptr, gnomekcfg);
    strcat(tmpptr, kanjidicstrg[i]);
    if (gconf_client_get_bool(gconf_client, tmpptr, NULL)) {
      conf.kdiccfg[i] = TRUE;
      /* printf("%s : %d\n",kanjidicstrg[i], conf.kdiccfg[i]); */
    }
    else conf.kdiccfg[i] = FALSE; 
    g_free(tmpptr);
  }
 
  //Load gjiten search history
  for (i = 0; i <= 50; i++) {
    sprintf(historystr, "/apps/gjiten/history/history%d", i);
    conf.history[i] = gconf_client_get_string(gconf_client, historystr, NULL);
    if (conf.history[i] == NULL) break;
  }
  
  if (conf.normalfont != NULL) {
    conf.normalfont_desc = pango_font_description_from_string(conf.normalfont);
  }

  if (conf.kanjidicfile == NULL) {
    conf.usekanjidic = FALSE;
  }
  else if (strlen(conf.kanjidicfile) == 0) conf.usekanjidic = FALSE;

  for (i = 1; i <= conf.numofdics; i++) {
    if (conf.dicfiles[i - 1] == NULL) {
      conf.numofdics = i - 1;
      break;
    }
  }
  
  //Check/Open Dictionary files
  conf.dicts_ok = TRUE;
  if (conf.usekanjidic) {
    conf.kanjidicfile_descriptor = open(conf.kanjidicfile, O_RDONLY);
    if (conf.kanjidicfile_descriptor == -1) {
      sprintf(gnome_dialog_msg, _("Error opening dictfile:  %s\nCheck your preferences!"), conf.kanjidicfile);
      //printf("kanjidic err\n");
      conf.dicts_ok = FALSE;
      return 2;
    }
  }

  for (i = 0; i < conf.numofdics; i++) { 
    if (conf.full_dicnames[i] != NULL) g_free(conf.full_dicnames[i]);
    conf.full_dicnames[i] = (char *) g_malloc(strlen(conf.dictpath) + strlen(conf.dicfiles[i]) + 1);
    sprintf(conf.full_dicnames[i], "%s%s", conf.dictpath, conf.dicfiles[i]);
    conf.dicfile_descriptor[i] = open(conf.full_dicnames[i], O_RDONLY);
    if (conf.dicfile_descriptor[i] == -1) {
      //printf("dic err %s: %d\n", fullfilename, conf.dicfile_descriptor[i]);
      sprintf(gnome_dialog_msg, _("Error opening dictfile:  %s\nCheck your preferences!"), conf.full_dicnames[i]);
      conf.dicts_ok = FALSE;
      return 3;
    }
  }
  return 0;
}

int save_settings() {
  gchar gnomekcfg[] = "/apps/gjiten/kanjidic/";
  gchar dicstr[50];
  int i;
  gchar *confpath;

  gconf_client_set_string(gconf_client, "/apps/gjiten/general/version", VERSION, NULL);
  //Save kanjidic display options
  for (i = 0; i < KCFGNUM; i++) { 
    confpath = g_malloc(strlen(gnomekcfg) + strlen(kanjidicstrg[i]) + 1);
    strcpy(confpath, gnomekcfg);
    strcat(confpath, kanjidicstrg[i]);
    gconf_client_set_bool(gconf_client, confpath, conf.kdiccfg[i], NULL);
    g_free(confpath);
  }
  
  //gconf_client_set_bool(gconf_client, "/apps/gjiten/general/tooltips", conf.tooltips, NULL);
  gconf_client_set_bool(gconf_client, "/apps/gjiten/general/menubar", conf.menubar, NULL);
  gconf_client_set_bool(gconf_client, "/apps/gjiten/general/toolbar", conf.toolbar, NULL);
  gconf_client_set_bool(gconf_client, "/apps/gjiten/general/usekanjipad", conf.usekanjipad, NULL);
  gconf_client_set_bool(gconf_client, "/apps/gjiten/general/usekanjidic", conf.usekanjidic, NULL);
  gconf_client_set_string(gconf_client, "/apps/gjiten/general/dictpath", conf.dictpath, NULL);
  gconf_client_set_string(gconf_client, "/apps/gjiten/kanjidic/kanjidicfile", conf.kanjidicfile, NULL);
  gconf_client_set_string(gconf_client, "/apps/gjiten/general/kanjipad", conf.kanjipad, NULL);
  gconf_client_set_int(gconf_client, "/apps/gjiten/general/numofdics", conf.numofdics, NULL);

  gconf_client_set_bool(gconf_client, "/apps/gjiten/general/bigwords", conf.bigwords, NULL);
  gconf_client_set_bool(gconf_client, "/apps/gjiten/general/bigkanji", conf.bigkanji, NULL);
  gconf_client_set_string(gconf_client, "/apps/gjiten/general/bigfont", conf.bigfont, NULL);
  gconf_client_set_string(gconf_client, "/apps/gjiten/general/normalfont", conf.normalfont, NULL);
  gconf_client_set_bool(gconf_client, "/apps/gjiten/general/gdk_use_xft", conf.gdk_use_xft, NULL);
  gconf_client_set_bool(gconf_client, "/apps/gjiten/general/force_ja_JP", conf.force_ja_JP, NULL);
  gconf_client_set_bool(gconf_client, "/apps/gjiten/general/force_language_c", conf.force_language_c, NULL);

  //Save dicfiles
  for (i = 0; i < conf.numofdics; i++) {
    if (i == MAXDICFILES - 1) break;
    sprintf(dicstr, "/apps/gjiten/general/dic%i", i);
    gconf_client_set_string(gconf_client, dicstr, conf.dicfiles[i], NULL);
  }
  return 1;
}

int save_history(GList *history) {
  char historystr[40];
  int i;
  if (history != NULL) {
    for (i = 0; i <= 50; i++) {
      sprintf(historystr,"/apps/gjiten/history/history%d", i);
      //printf("%s: %s\n", historystr, (char*) history->data);
      gconf_client_set_string(gconf_client, historystr, history->data, NULL);
      history = g_list_next(history);
      if (history == NULL) break;
    } 
    return 1;
  }
  else return 0;
}

void save_options() {
    gconf_client_set_bool(gconf_client, "/apps/gjiten/general/autoadjust_enabled", conf.autoadjust_enabled, NULL);
    gconf_client_set_bool(gconf_client, "/apps/gjiten/general/deinflection_enabled", conf.deinflection_enabled, NULL);
    gconf_client_set_bool(gconf_client, "/apps/gjiten/general/searchlimit_enabled", conf.searchlimit_enabled, NULL);
    gconf_client_set_int(gconf_client, "/apps/gjiten/general/maxwordmatches", conf.maxwordmatches, NULL);
}

