/* gjiten.c

   GJITEN : A GTK+/GNOME BASED JAPANESE DICTIONARY
  
   Copyright (C) 1999 - 2003 Botond Botyanszki <boti@rocketmail.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published  by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <libgnome/libgnome.h>
#include <libgnomeui/libgnomeui.h>
#include <gconf/gconf-client.h>
#include <stdint.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
//#include <locale.h>

#include "gjiten.h"
#include "kanjidic.h"
#include "gjiten-settings.h"

#define LINBUFFSIZE 1024*60
#define MAX_MATCHES conf.maxwordmatches

#define USE_MMAP

/*====== Prototypes========================================================*/
void create_about();
void display_manual();
extern void create_window_kanjidic();
extern void set_strginfo();
static gchar *get_eof_line(gchar *ptr, gchar *end_ptr);
void create_window_gjiten();
void start_kanjipad();
extern void create_dialog_preferences();
void gjiten_search(unsigned char *srchstrg);
void gjiten_paste();
void gjiten_copy();
static void gjiten_exit();
void on_text_entered();
extern void load_radkfile();
void linear_search(int srchtype, int dic_no, int32_t index_pos, int srchstrg_len, unsigned char *srchstrg,
		   int *srch_resp, int32_t *res_index, int *hit_pos, int *res_len, unsigned char *res_str);
void search4string(int type, int dic_no, int32_t index_pos, int srchstrg_len, unsigned char *srchstrg,
		   int *srch_resp, int32_t *res_index, int *hit_pos, int *res_len, unsigned char *res_str);
extern void print_kanjinfo(gunichar kanji);
void AppKanji(unsigned char c1,unsigned char c2);
static void Verbinit();
void KanjRad(unsigned char *kanjistr);
static int isKanaChar(gunichar c);
//static int isKatakanaChar(gunichar c);
//static int isHiraganaChar(gunichar c);
static int isKanjiChar(gunichar c);
static void parse_an_arg(poptContext state,
			 enum poptCallbackReason reason,
			 const struct poptOption *opt,
			 const char *arg, void *data);
void on_back_clicked();
void on_forward_clicked();
int get_jp_match_type(gchar *line, gchar *srchstrg, int offset);
static void print_result(gchar *txt2print, int result_offset, gchar *searchstrg);

/*====== end of prototypes==================================================*/


/***************** VARIABLES ***********************/

int i; 
gchar *gnome_dialog_msg = NULL;
gchar *clipboard_text = NULL;
int shutdown_gjiten = TRUE;
int append_to_history = TRUE;
int  engsrch, jpsrch;
gchar rcpath[110];
int default_flag;
int word_matches;
gchar *linsrchbuff;
gchar line_kanji[400], line_reading[400];
uint32_t srchpos;
int dicname_printed;

/* GTK variables */
GtkWidget *window_gjiten;
extern GtkWidget *window_kanjidic;
extern GtkWidget *dialog_preferences;
GtkWidget *combo_entry;
GList *combo_entry_glist = NULL;
GtkWidget *text_results_view;
GtkTextBuffer *text_results_buffer;
GtkWidget *menu_selectdic;
GtkWidget *combo_entry_dictfile;
GtkWidget *checkb_verb;
GtkWidget *checkb_autoadjust;
GtkWidget *checkb_searchlimit;
GtkWidget *spinb_searchlimit;
GtkWidget *radiob_jpexact;
GtkWidget *radiob_startw;
GtkWidget *radiob_endw;
GtkWidget *radiob_any;
GtkWidget *radiob_engexact;
GtkWidget *radiob_words;
GtkWidget *radiob_partial;
GtkWidget *radiob_searchdic;
GtkWidget *radiob_searchall;
GtkWidget *button_back;
GtkWidget *button_forward;
GdkColor color1;
GdkColor color2;
GdkColormap *cmap;
GtkTextIter iter;
GtkWidget *appbar_mainwin;
gpointer current_glist_word = NULL;
gchar *kanjidicstrg[] = { "kanji", "radical", "strokes", "reading", "korean",
			  "pinyin", "english", "bushu", "classic", "freq", "jouyou",
			  "deroo", "skip", "fourc", "hindex", "nindex", "vindex",
			  "iindex", "mnindex", "mpindex", "eindex", "kindex", 
			  "lindex", "oindex", "cref", "missc", "unicode", "jisascii" };


enum {
  KANJIDIC_KEY        = -1,
  WORD_LOOKUP_KEY     = -2,
  KANJI_LOOKUP_KEY    = -3,
  CLIP_KANJI_KEY      = -4,
  CLIP_WORD_KEY       = -5
};

/* Command line arguments via popt */
static struct poptOption arg_options [] = {
  { NULL, '\0', POPT_ARG_CALLBACK, (gpointer)parse_an_arg, 0},

  { "kanjidic", 'k', POPT_ARG_NONE, NULL, KANJIDIC_KEY, 
    N_("Start up Kanjidic instead of Word dictionary"), NULL},
 
  { "word-lookup", 'w', POPT_ARG_STRING, NULL, WORD_LOOKUP_KEY,
    N_("Look up WORD in first dictionary"), N_("WORD") },
   
  { "kanji-lookup", 'l', POPT_ARG_STRING, NULL, KANJI_LOOKUP_KEY,
    N_("Look up KANJI in kanji dictionary"), N_("KANJI") },

  { "clip-kanji", 'c', POPT_ARG_NONE, NULL, CLIP_KANJI_KEY,
    N_("Look up kanji from clipboard"), NULL },

  { "clip-word", 'v', POPT_ARG_NONE, NULL, CLIP_WORD_KEY,
    N_("Look up word from clipboard"), NULL },

  { NULL, '\0', 0, NULL, 0}
  
};


static GnomeUIInfo file_menu_uiinfo[] =
{
  GNOMEUIINFO_MENU_EXIT_ITEM(GTK_SIGNAL_FUNC(gjiten_exit), NULL),
  GNOMEUIINFO_END
};

static GnomeUIInfo edit_menu_uiinfo[] =
{
  //GNOMEUIINFO_MENU_CUT_ITEM(NULL, NULL),
  GNOMEUIINFO_MENU_COPY_ITEM(gjiten_copy, NULL),
  GNOMEUIINFO_MENU_PASTE_ITEM(gjiten_paste, NULL),
  // GNOMEUIINFO_MENU_CLEAR_ITEM(clear_entry_box, GTK_OBJECT(GTK_COMBO(combo_entry)->entry)),
  GNOMEUIINFO_SEPARATOR,
  GNOMEUIINFO_MENU_PREFERENCES_ITEM(create_dialog_preferences, NULL),
  GNOMEUIINFO_END
};

static GnomeUIInfo tools_menu_uiinfo[] =
{
  {
    GNOME_APP_UI_ITEM, N_("KanjiDic"), NULL, create_window_kanjidic, 
    NULL, NULL, GNOME_APP_PIXMAP_NONE, NULL, 0, 0, NULL
  },
  {
    GNOME_APP_UI_ITEM, N_("KanjiPad"), NULL, start_kanjipad, NULL, NULL,
    GNOME_APP_PIXMAP_NONE, NULL, 0, 0, NULL
  },
  GNOMEUIINFO_END
};

static GnomeUIInfo help_menu_uiinfo[] =  {
  {
    GNOME_APP_UI_ITEM, N_("_Manual"), N_("Display the Gjiten Manual"),
    display_manual, NULL, NULL,
    GNOME_APP_PIXMAP_STOCK, GTK_STOCK_HELP,
    0, 0, NULL },
  
  {    
    GNOME_APP_UI_ITEM, N_("_About"), N_("Information about the program"),
    create_about, NULL, NULL,
    GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_ABOUT,
    0, 0, NULL },
  
  GNOMEUIINFO_END
};

static GnomeUIInfo menubar_uiinfo[] =
  {
    GNOMEUIINFO_MENU_FILE_TREE(file_menu_uiinfo),
    GNOMEUIINFO_MENU_EDIT_TREE(edit_menu_uiinfo),
    {
      GNOME_APP_UI_SUBTREE, N_("Tools"), NULL, tools_menu_uiinfo, NULL, 
      NULL,  GNOME_APP_PIXMAP_NONE, N_("Tools"), 0, 0, NULL
    },
    GNOMEUIINFO_MENU_HELP_TREE(help_menu_uiinfo),
    GNOMEUIINFO_END
  };

/*
static GnomeUIInfo toolbar_uuinfo[] =  {
  GNOMEUIINFO_ITEM_STOCK("Exit", N_("Exit Gjiten"), gjiten_exit, GNOME_STOCK_PIXMAP_EXIT),
  GNOMEUIINFO_ITEM_STOCK("Back", N_("Back"), on_back_clicked, GNOME_STOCK_PIXMAP_BACK),
  //    gtk_widget_set_sensitive(button_back, FALSE);
  GNOMEUIINFO_ITEM_STOCK("Forward", N_("Forward"), on_forward_clicked, GNOME_STOCK_PIXMAP_FORWARD),
  GNOMEUIINFO_ITEM_STOCK("KanjiDic", N_("KanjiDic"), create_window_kanjidic, GNOME_STOCK_PIXMAP_BOOK_BLUE),
  //    gtk_widget_set_sensitive(button_kanjidic,conf.usekanjidic);
  GNOMEUIINFO_ITEM_STOCK("KanjiPad", N_("KanjiPad"), start_kanjipad, GNOME_STOCK_PIXMAP_PREFERENCES),
  //    gtk_widget_set_sensitive(button_kanjipad,conf.usekanjipad);
  GNOMEUIINFO_ITEM_STOCK("Search", N_("Search"), on_text_entered, GNOME_STOCK_PIXMAP_SEARCH),
  // GNOMEUIINFO_ITEM_STOCK("Copy", N_("Copy"), gjiten_copy, GNOME_STOCK_PIXMAP_COPY),
  // GNOMEUIINFO_ITEM_STOCK("Paste", N_("Paste"), gjiten_paste, GNOME_STOCK_PIXMAP_PASTE),
  GNOMEUIINFO_END 
};
*/

// xjdic variables
//int jiver = 14;		// The last time the index structure changed was Version1.4 
//extern int NoDics;
//extern uint32_t dichits[40],dicmiss[40];
//extern uint32_t indhits[40],indmiss[40];
//extern uint32_t vbkills;
//char *DicDir;    // PATH TO DICTIONARY FILES
//int verblen;
//int xfilelen;
//int DicNum;

//verb deinflection
struct vinfl_struct {
  gchar *conj;
  gchar *infl;
  gchar *type;
};
gchar *vconj_types[40];
GSList *vinfl_list = NULL;

/*================ Functions ===============================================*/

static void parse_an_arg(poptContext state,
			 enum poptCallbackReason reason,
			 const struct poptOption *opt,
			 const char *arg, void *data) {

  
  switch (opt->val) {
  case KANJIDIC_KEY:
    conf.startkanjidic = TRUE;
    break;
  case WORD_LOOKUP_KEY:
    conf.word_to_lookup = (gchar *)arg;
    break;
  case KANJI_LOOKUP_KEY:
    conf.kanji_to_lookup = (gchar *)arg;
    break;
  case CLIP_KANJI_KEY:
    conf.clip_kanji_lookup = TRUE;
    conf.clip_word_lookup = FALSE;
    break;
  case CLIP_WORD_KEY:
    conf.clip_word_lookup = TRUE;
    conf.clip_kanji_lookup = FALSE;
    break;
  default:
    break;
  }
}

void gjiten_copy() {
  gchar *selection = NULL;

  selection = gtk_clipboard_wait_for_text(gtk_clipboard_get(GDK_SELECTION_PRIMARY));
  if (selection == NULL) return;
  gtk_clipboard_set_text(gtk_clipboard_get(GDK_SELECTION_CLIPBOARD), selection, -1);
}

 
void gjiten_paste() {
  gchar *selection = NULL;

  // First try the current selection
  selection = gtk_clipboard_wait_for_text(gtk_clipboard_get(GDK_SELECTION_PRIMARY));
  if (selection != NULL) gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(combo_entry)->entry), selection);
  else {
    // if we didn't get anything, try the default clipboard
    selection = gtk_clipboard_wait_for_text(gtk_clipboard_get(GDK_SELECTION_CLIPBOARD));
      if (selection != NULL) gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(combo_entry)->entry), selection);
  }
}

void show_errors() {
  GtkWidget *dialog = NULL;
  if (strlen(gnome_dialog_msg)) {  /* display error msg */ 
    dialog = gtk_message_dialog_new(GTK_WINDOW(window_gjiten),
				    GTK_DIALOG_DESTROY_WITH_PARENT,
				    GTK_MESSAGE_ERROR,
				    GTK_BUTTONS_OK,
				    "%s",
				    gnome_dialog_msg);
    gtk_dialog_run(GTK_DIALOG(dialog));
    if (dialog != NULL) gtk_widget_destroy(dialog);
    dialog = NULL;
    gnome_dialog_msg[0] = 0;
  }
}

void exit_critical(char *exitmsg) { /*gnomeify this.*/

  g_print("%s\nExiting.\n",exitmsg);
  exit(1);
}


static void gjiten_exit() {
  // printf("gjiten_exit\n");
  if (shutdown_gjiten) {
    save_history(g_list_first(combo_entry_glist));
    save_options();
    gtk_widget_destroy(window_gjiten);
    window_gjiten = NULL;
    if (window_kanjidic == NULL) {
      g_free(gnome_dialog_msg); 
#ifndef USE_MMAP
      if (linsrchbuff) g_free(linsrchbuff);
#endif
      if (cmap) g_free(cmap);
      exit(0);
    }
  }
}

void start_kanjipad() {
  FILE *kanjipad_binary;
  char *kpad_cmd;

  kanjipad_binary = fopen(conf.kanjipad,"r");
  if (kanjipad_binary == NULL) {
    sprintf(gnome_dialog_msg, (_("Couldn't find the KanjiPad executable!\n"
				 "Please make sure you have it installed on your system \n"
				 "and set the correct path to it in the Preferences.\n"
				 "See the Documentation for more details about KanjiPad.")));
    show_errors();
  }
  else {
    fclose(kanjipad_binary);
    kpad_cmd = g_malloc(strlen(conf.kanjipad)+2);
    strcpy(kpad_cmd,conf.kanjipad);
    strcat(kpad_cmd,"&");
    system(kpad_cmd);
    g_free(kpad_cmd);
  }
}

// Compares strg1 with strg2.
// If strg1 == strg3|strg2 then returns TRUE (End of strg1 matches strg2)
static int strg_end_compare(gchar *strg1, gchar *strg2) {
  int i = 0;
  int matching = TRUE;
  gchar *strg1_end, *strg2_end;

  if (strlen(strg1) < strlen(strg2)) return FALSE;

  strg1_end = strg1 + strlen(strg1);
  strg2_end = strg2 + strlen(strg2);

  for (i = 0; i < g_utf8_strlen(strg2, -1); i++) {
    strg1_end = g_utf8_prev_char(strg1_end);
    strg2_end = g_utf8_prev_char(strg2_end);
    if (g_utf8_get_char(strg1_end) != g_utf8_get_char(strg2_end)) 
      matching = FALSE;
  }
  return matching;
}

static gchar *get_eof_line(gchar *ptr, gchar *end_ptr) {
  static gchar *tmpptr;
  tmpptr = ptr;
  while (*tmpptr != '\n') {
    if (end_ptr == tmpptr) return NULL;
    tmpptr++;
  }
  tmpptr++;
  return tmpptr;
}

// Load & initialize verb inflection details
static void Verbinit() {
  static int verbinit_done = FALSE;
  gchar *tmp_ptr;
  int vinfl_size = 0;
  struct stat vinfl_stat;
  gchar *vinfl_start, *vinfl_ptr, *vinfl_end;
  int error = FALSE;
  int fd = 0;
  int vinfl_part = 1;
  int conj_type = 40;
  struct vinfl_struct *tmp_vinfl_struct;
  GSList *tmp_list_ptr = NULL;

  if (verbinit_done == TRUE) {
    //printf("Verbinit already done!\n");
    return;
  }

  if (stat(VINFL_FILENAME, &vinfl_stat) != 0) {
    printf("**ERROR** %s: stat() \n", VINFL_FILENAME);
    error = TRUE;
  }
  vinfl_size = vinfl_stat.st_size;
  fd = open(VINFL_FILENAME, O_RDONLY);
  if (fd == -1) {
    printf("**ERROR** %s: open()\n", VINFL_FILENAME);
    error = TRUE;
  }
  // printf("SIZE: %d\n", radkfile_size);
  vinfl_start = (gchar *) mmap(NULL, vinfl_size, PROT_READ, MAP_SHARED, fd, 0);
  if (vinfl_start == NULL) exit_critical("mmap() failed for "VINFL_FILENAME"\n");

  //  printf("STRLEN: %d\n", strlen(radkfile));

  if (error == TRUE) {
    if (dialog_preferences == NULL) {
      sprintf(gnome_dialog_msg, _("Error opening %s.\n "\
				  "Check your preferences or read the documentation!"),
	      VINFL_FILENAME); 
      show_errors();
    }
    conf.dicts_ok = FALSE;
    return;
  }

  vinfl_end = vinfl_start + strlen(vinfl_start);
  vinfl_ptr = vinfl_start;

  vinfl_part = 1;
  while ((vinfl_ptr < vinfl_end) && (vinfl_ptr != NULL)) {
    if (*vinfl_ptr == '#') {  //find $ as first char on the line
      vinfl_ptr = get_eof_line(vinfl_ptr, vinfl_end); //Goto next line
      continue;
    }
    if (*vinfl_ptr == '$') vinfl_part = 2;
    
    switch (vinfl_part) {
    case 1:
      if (g_ascii_isdigit(*vinfl_ptr) == TRUE) { //Conjugation numbers
	conj_type = atoi(vinfl_ptr);
	if ((conj_type < 0) || (conj_type > 39)) break;
	while (g_ascii_isdigit(*vinfl_ptr) == TRUE) vinfl_ptr = g_utf8_next_char(vinfl_ptr); //skip the number
	while (g_ascii_isspace(*vinfl_ptr) == TRUE) vinfl_ptr = g_utf8_next_char(vinfl_ptr); //skip the space
	tmp_ptr = vinfl_ptr; // beginning of conjugation definition;
	vinfl_ptr = get_eof_line(vinfl_ptr, vinfl_end); //find end of line
	vconj_types[conj_type] = g_strndup(tmp_ptr, 
					   vinfl_ptr - tmp_ptr -1);
	//printf("%d : %s\n", conj_type, vconj_types[conj_type]);
      }
      break;
    case 2:
      if (g_unichar_iswide(g_utf8_get_char(vinfl_ptr)) == FALSE) {
	vinfl_ptr =  get_eof_line(vinfl_ptr, vinfl_end);
	break;
      }
      tmp_vinfl_struct = (struct vinfl_struct *) malloc (sizeof(struct vinfl_struct));
      tmp_ptr = vinfl_ptr;
      while (g_unichar_iswide(g_utf8_get_char(vinfl_ptr)) == TRUE) 
	vinfl_ptr = g_utf8_next_char(vinfl_ptr); //skip the conjugation
      tmp_vinfl_struct->conj = g_strndup(tmp_ptr, vinfl_ptr - tmp_ptr); //store the conjugation
      while (g_ascii_isspace(*vinfl_ptr) == TRUE) vinfl_ptr = g_utf8_next_char(vinfl_ptr); //skip the space
      tmp_ptr = vinfl_ptr;
      while (g_unichar_iswide(g_utf8_get_char(vinfl_ptr)) == TRUE) 
	vinfl_ptr = g_utf8_next_char(vinfl_ptr); //skip the inflection
      tmp_vinfl_struct->infl = g_strndup(tmp_ptr, vinfl_ptr - tmp_ptr); //store the inflection
      while (g_ascii_isspace(*vinfl_ptr) == TRUE) vinfl_ptr = g_utf8_next_char(vinfl_ptr); //skip the space
      tmp_vinfl_struct->type = vconj_types[atoi(vinfl_ptr)];
      vinfl_ptr =  get_eof_line(vinfl_ptr, vinfl_end);
      //printf("%s|%s|%s\n", tmp_vinfl_struct->conj, tmp_vinfl_struct->infl, tmp_vinfl_struct->type);
      tmp_list_ptr = g_slist_append(tmp_list_ptr, tmp_vinfl_struct);
      if (vinfl_list == NULL) vinfl_list = tmp_list_ptr;
      break;
    }
  } 
  verbinit_done = TRUE;
}

static void get_kanji_and_reading(gchar *line) {
  gchar *currentchar;
  gchar *reading_start;

  currentchar = line;

  while (g_unichar_isspace(*currentchar) == FALSE) { // find first space
    if (currentchar - line >= strlen(line)) break;
    currentchar = g_utf8_next_char(currentchar);
  }
  if ((*(currentchar+1)) == '/') { //only have kana
    line_kanji[0] = 0;
    strncpy(line_reading, line, currentchar - line);
    line_reading[currentchar - line] = 0;
  }
  else { //copy both kanji and reading
    strncpy(line_kanji, line, currentchar - line);
    line_kanji[currentchar - line] = 0;
    currentchar +=2;
    reading_start = currentchar;
    while (g_unichar_isspace(*currentchar) == FALSE) { // find first space
      if (currentchar - line >= strlen(line)) break;
      currentchar = g_utf8_next_char(currentchar);
    }
    currentchar--;
    strncpy(line_reading, reading_start, currentchar - reading_start);
    line_reading[currentchar - reading_start] = 0;
  }
}

void print_verb_inflections(gchar *srchstrg) {
  int srchresp, roff, rlen;
  int32_t oldrespos, respos;
  int gjit_search = SRCH_START;
  unsigned char repstr[1024];
  GSList *tmp_list_ptr;
  struct vinfl_struct *tmp_vinfl_struct;
  gchar *deinflected, *currentchar, *prevresult;
  int printit = TRUE;

  tmp_list_ptr = vinfl_list;
  if (vinfl_list == NULL) { 
    printf("VINFL LIST == NULL\n");
    return;
  }
  
  deinflected = (gchar *) g_malloc(strlen(srchstrg) + 20);
  
  do {
    tmp_vinfl_struct = (struct vinfl_struct *) tmp_list_ptr->data;
    if (strg_end_compare(srchstrg, tmp_vinfl_struct->conj) == TRUE) {

      /*
      gtk_text_buffer_insert(GTK_TEXT_BUFFER(text_results_buffer), &iter, tmp_vinfl_struct->type, -1);
      gtk_text_buffer_insert(GTK_TEXT_BUFFER(text_results_buffer), &iter, ": ", -1);
      gtk_text_buffer_insert(GTK_TEXT_BUFFER(text_results_buffer), &iter, tmp_vinfl_struct->conj, -1);
      gtk_text_buffer_insert(GTK_TEXT_BUFFER(text_results_buffer), &iter, "->", -1);
      gtk_text_buffer_insert(GTK_TEXT_BUFFER(text_results_buffer), &iter, tmp_vinfl_struct->infl, -1);
      */
      // create deinflected string
      strncpy(deinflected, srchstrg, strlen(srchstrg) - strlen(tmp_vinfl_struct->conj));
      strcpy(deinflected + strlen(srchstrg) - strlen(tmp_vinfl_struct->conj), 
	      tmp_vinfl_struct->infl);

      /*
      gtk_text_buffer_insert(GTK_TEXT_BUFFER(text_results_buffer), &iter, " =>", -1);
      gtk_text_buffer_insert(GTK_TEXT_BUFFER(text_results_buffer), &iter, deinflected, -1);
      gtk_text_buffer_insert(GTK_TEXT_BUFFER(text_results_buffer), &iter, "\n", -1);
      */
      oldrespos = srchpos = 0;    
      gjit_search = SRCH_START;
      prevresult = NULL;
      do { // search loop
	oldrespos = respos;
    
	search4string(gjit_search, conf.selected_dic , srchpos, strlen(deinflected), deinflected, 
		      &srchresp, &respos, &roff, &rlen, repstr);
	//    printf("respos:%d, oldrespos:%d, roffset:%d, rlen:%d\nrepstr:%s\n", respos, oldrespos, roff, rlen, repstr);
	if (srchresp != SRCH_OK)  {
	  break;   //No more matches
	}
	if (gjit_search == SRCH_START) {
	  srchpos = respos;
	  gjit_search = SRCH_CONT;
	}
	srchpos++;
	if (oldrespos == respos) continue;

	printit = TRUE;
	//Check if japanese word is only kanjis
	get_kanji_and_reading(repstr);
	currentchar = g_utf8_prev_char(line_kanji + strlen(line_kanji));
	while (isKanaChar(g_utf8_get_char(currentchar))) {
	  if (currentchar == line_kanji) break;
	  currentchar = g_utf8_prev_char(currentchar);
	}
	if (currentchar == g_utf8_prev_char(line_kanji + strlen(line_kanji))) printit = FALSE;
	if (strlen(tmp_vinfl_struct->conj) == strlen(srchstrg)) 
	  printit = FALSE; // don't display if conjugation is the same length as th srchstrg
	if (get_jp_match_type(repstr, deinflected, roff) != EXACT_MATCH) 
	  printit = FALSE; // Display only EXACT_MATCHes

	if (printit == TRUE) {
	  gtk_text_buffer_insert_with_tags_by_name(GTK_TEXT_BUFFER(text_results_buffer), &iter,  
						   _("Possible inflected verb or adjective: "), 
						   -1, "brown_foreground", NULL);
	  gtk_text_buffer_insert(GTK_TEXT_BUFFER(text_results_buffer), &iter, tmp_vinfl_struct->type, -1);
	  gtk_text_buffer_insert(GTK_TEXT_BUFFER(text_results_buffer), &iter, " " , -1);
	  gtk_text_buffer_insert(GTK_TEXT_BUFFER(text_results_buffer), &iter, tmp_vinfl_struct->conj, -1);
	  gtk_text_buffer_insert(GTK_TEXT_BUFFER(text_results_buffer), &iter, "->", -1);
	  gtk_text_buffer_insert(GTK_TEXT_BUFFER(text_results_buffer), &iter, tmp_vinfl_struct->infl, -1);
	  gtk_text_buffer_insert(GTK_TEXT_BUFFER(text_results_buffer), &iter, "\n   ", -1);
	  print_result(repstr, roff, deinflected);
	  word_matches++;
	}
      } while (srchresp == SRCH_OK);
    }
  }    
  while ((tmp_list_ptr = g_slist_next(tmp_list_ptr)) != NULL);
  
  g_free(deinflected);

}

static void print_result(gchar *txt2print, int result_offset, gchar *searchstrg) {
  gchar *strg_to_roff;
  glong strlen_to_roff;
  gchar *currentchar;
  //gchar left_bracket[3]; 
  //gchar right_bracket[3];
  gchar *kana_start;
  gchar *exp_start;
  GtkTextMark *linestart;
  GtkTextIter startmatch, endmatch;

  linestart = gtk_text_buffer_create_mark(GTK_TEXT_BUFFER(text_results_buffer), "linestart", &iter, TRUE);
  /*
  left_bracket[0] = 0xEF;
  left_bracket[1] = 0xBC;
  left_bracket[2] = 0x88;
  right_bracket[0] = 0xEF;
  right_bracket[1] = 0xBC;
  right_bracket[2] = 0x89;
  */

  strg_to_roff = (gchar *) g_strndup(txt2print, result_offset);
  strlen_to_roff = g_utf8_strlen(strg_to_roff, -1);
  
  currentchar = txt2print;

  while (g_unichar_isspace(*currentchar) == FALSE) { // find first space
    if (currentchar - txt2print >= strlen(txt2print)) break;
    currentchar = g_utf8_next_char(currentchar);
  }
  
  //print out japanese word
  if (conf.bigwords == FALSE) {
    gtk_text_buffer_insert(GTK_TEXT_BUFFER(text_results_buffer), &iter, txt2print, currentchar - txt2print); 
  }
  else {
    gtk_text_buffer_insert_with_tags_by_name(GTK_TEXT_BUFFER(text_results_buffer), &iter, txt2print, 
					     currentchar - txt2print, "bigfont", NULL);
  }
  
  currentchar = g_utf8_next_char(currentchar);
  if (*(currentchar) == '[')  { //we have a kana reading
    //gtk_text_buffer_insert(GTK_TEXT_BUFFER(text_results_buffer), &iter, left_bracket, 3);
    gtk_text_buffer_insert(GTK_TEXT_BUFFER(text_results_buffer), &iter, " (", 2);
    currentchar = kana_start = g_utf8_next_char(currentchar);
    while(*(currentchar) != ']') { //find ending ]
      currentchar = g_utf8_next_char(currentchar);
    }
    gtk_text_buffer_insert(GTK_TEXT_BUFFER(text_results_buffer), &iter, kana_start, currentchar - kana_start); //print out kana reading
    //gtk_text_buffer_insert(GTK_TEXT_BUFFER(text_results_buffer), &iter, right_bracket, 3); 
    gtk_text_buffer_insert(GTK_TEXT_BUFFER(text_results_buffer), &iter, ") ", 2);
    currentchar += 3;
  }
  else {
    gtk_text_buffer_insert(GTK_TEXT_BUFFER(text_results_buffer), &iter, " ", 1); // insert space 
    currentchar++;
  }

  while ((currentchar < txt2print + strlen(txt2print)) && (*currentchar != '\n')) {
    exp_start = currentchar;
    while ((*currentchar != '/')) { 
      currentchar = g_utf8_next_char(currentchar);  
    }
    gtk_text_buffer_insert(GTK_TEXT_BUFFER(text_results_buffer), &iter, exp_start, currentchar - exp_start); //print out expression
    gtk_text_buffer_insert(GTK_TEXT_BUFFER(text_results_buffer), &iter, "; ", 2);
    currentchar++;
  }
  gtk_text_buffer_insert(GTK_TEXT_BUFFER(text_results_buffer), &iter, "\n", 1); // insert linebreak

  //find searchstrg matches in the line we print and highlight it.

  gtk_text_buffer_get_iter_at_mark(GTK_TEXT_BUFFER(text_results_buffer), &endmatch, linestart);
  
  while (gtk_text_iter_forward_search(&endmatch, searchstrg, 0, &startmatch, &endmatch, &iter) == TRUE) 
    gtk_text_buffer_apply_tag_by_name(GTK_TEXT_BUFFER(text_results_buffer), 
				      "blue_foreground", &startmatch, &endmatch);
  
} 

// this is a chooser between search functions: linear and indexed
void search4string(int type, int dic_no, int32_t index_pos, int srchstrg_len, unsigned char *srchstrg,
		   int *srch_resp, int32_t *res_index, int *hit_pos, int *res_len, unsigned char *res_str) { 
  linear_search(type, dic_no, index_pos, srchstrg_len, srchstrg, srch_resp, 
		res_index, hit_pos, res_len, res_str);
}

void linear_search(int srchtype, int dic_no, int32_t index_pos, int srchstrg_len, unsigned char *srchstrg,
		   int *srch_resp, int32_t *res_index, int *hit_pos, int *res_len, unsigned char *res_str) {
  char *linestart, *lineend; 
  int copySize = 1023;
  static char *linsrchptr;
  int filedescr;
  char *dicname;
#ifdef USE_MMAP
  struct stat dicfile_stat;
  static int dicfile_size;
#else
  static int bytesread, readnew;
  int is_file_start = TRUE;
  int memsrchlen = 0;
#endif

  if (dic_no == -1) {
    filedescr = conf.kanjidicfile_descriptor;
    dicname = conf.kanjidicfile;
  }
  else {
    filedescr = conf.dicfile_descriptor[dic_no];
    dicname = conf.full_dicnames[dic_no];
  }

#ifdef USE_MMAP
  if ((dic_no != conf.mmaped_dic_num) || (conf.mmaped_dic == NULL)) {
    //free mem of previously used dicfile
    if ((conf.mmaped_dic_num != -2) && (conf.mmaped_dic != NULL)) {
      munmap(conf.mmaped_dic, dicfile_size);
    }
    
    //mmap dicfile into memory
    if (stat(dicname, &dicfile_stat) != 0) {
      printf("**ERROR** %s: stat() \n", dicname);
    }
    dicfile_size = dicfile_stat.st_size;
    conf.mmaped_dic = (gchar *) mmap(NULL, dicfile_size, PROT_READ, MAP_SHARED, filedescr, 0);
    if (conf.mmaped_dic == NULL) exit_critical("mmap() failed\n");
    conf.mmaped_dic_num = dic_no;
  }

  if (srchtype == SRCH_START) {
    linsrchptr = conf.mmaped_dic;
  }
 bad_hit:
  *srch_resp = SRCH_FAIL; // assume search fails 
  linsrchptr = strstr(linsrchptr, srchstrg);
  if (linsrchptr) {  // if we have a match
    linestart = linsrchptr;
    while ((*linestart != '\n') && (linestart != conf.mmaped_dic)) linestart--; // find beginning of line
    if (linestart == conf.mmaped_dic) {   
      if ((isKanjiChar(g_utf8_get_char(linestart)) == FALSE) && (isKanaChar(g_utf8_get_char(linestart)) == FALSE)) {
	linsrchptr++;
	goto bad_hit;
      }
    }

    linestart++;
    lineend = linestart;
    *hit_pos = linsrchptr - linestart;
    while (*lineend != '\n') { // find end of line
      lineend++;
      if (lineend >= conf.mmaped_dic + dicfile_size) { 
	printf("weird.\n");
	break;
      }
    }
    linsrchptr++;	
    if ((lineend - linestart + 1) < 1023)
      copySize = lineend - linestart + 1;
    else copySize = 1023;
    strncpy(res_str, linestart, copySize);
    res_str[copySize] = 0;
    //    printf("%s\n", res_str);
    (*res_index)++; // fool search_dict() 
    *srch_resp = SRCH_OK; // search succeeded 
  }
  return;

#else //NON-MMAP SEARCH

  *srch_resp = SRCH_FAIL; // assume search fails 
  if (srchtype == SRCH_START) {
    is_file_start = TRUE;
    lseek(filedescr, 0, SEEK_SET); // reset dicfile if new search
    readnew = TRUE; // and read new buff
  }
  
  do {
    if (readnew) {
      bytesread = read(filedescr, linsrchbuff, LINBUFFSIZE); // read a chunk into buffer
      //printf("Read %d bytes.\n",bytesread);
      readnew = FALSE;
      if (bytesread == 0) break; // read failed
      memsrchlen = bytesread;
      linsrchptr = linsrchbuff; // start from beginning of buffer
    }
    
    linsrchptr = strstr(linsrchptr, srchstrg);
    if (linsrchptr) {  // if we have a match
      linestart = linsrchptr;
      while ((*linestart != '\n') && (linestart != linsrchbuff)) linestart--; // find beginning of line
      if (linestart == linsrchbuff) {
	//printf("BAD LUCK. SEEKING BACK.\n");
	//we need to seek back.
	if (is_file_start == FALSE) {
	  while (TRUE) {
	    if (lseek(filedescr, -1000, SEEK_CUR) == -1) // seek back 1000 bytes
	      g_warning("seek error!!!\n"); 
	    read(filedescr, linsrchbuff, LINBUFFSIZE);
	    linestart = linsrchbuff + 1000;
	    linsrchptr += 1000;
	    while ((*linestart != '\n') && (linestart != linsrchbuff)) linestart--; // find beginning of line
	    if (linestart != linsrchbuff) break;
	  }
	}
      }
      linestart++;
      lineend = linestart;
      *hit_pos = linsrchptr - linestart;
      while (*lineend != '\n') { // find end of line
	lineend++;
	if (lineend >= linsrchbuff + bytesread) { 
	  if (lseek(filedescr, linestart - linsrchptr, SEEK_CUR) == -1) // seek back to beg of line
	    g_warning("seek error!\n"); 
	  readnew = TRUE;
	  break;
	}
      }
      if (readnew) continue;
      linsrchptr++;	
      if ((lineend - linestart + 1) < 1023)
	copySize = lineend - linestart + 1;
      else copySize = 1023;
      strncpy(res_str, linestart, copySize);
      res_str[copySize] = 0;
      (*res_index)++; // fool search_dict() 
      memsrchlen = bytesread + linsrchbuff - linsrchptr; 
      *srch_resp = SRCH_OK; // search succeeded 
      return;
    }
    else readnew = TRUE; // no match in buffer
    is_file_start = FALSE;
  } while (bytesread == LINBUFFSIZE);
#endif
}


//Finds out if the result is EXACT_MATCH, START_WITH_MATCH, END_WITH_MATCH, ANY_MATCH
int get_jp_match_type(gchar *line, gchar *srchstrg, int offset) {
  int srchstrglen;

  srchstrglen = strlen(srchstrg);
  if (offset == 0) { //can be EXACT or START_WITH
    if ((*(line + srchstrglen)) == ' ') return EXACT_MATCH;
    return START_WITH_MATCH;
  }
  else { //Check for Furigana
    if (g_unichar_isalpha(g_utf8_get_char(g_utf8_prev_char(line + offset))) == FALSE) {
      if (g_unichar_isalpha(g_utf8_get_char(line + offset + srchstrglen)) == FALSE) {
	return EXACT_MATCH;
      }
      else return START_WITH_MATCH;
    }
    else {
      if (g_unichar_isalpha(g_utf8_get_char(line + offset + srchstrglen)) == FALSE)
	return END_WITH_MATCH;
    }
  }
  if ((*(line + offset + srchstrglen)) == ' ') return END_WITH_MATCH;
  return ANY_MATCH;
}


  // According to the Unicode spec, char ranges are:
  // Kanji (19968 - 40959)
  // Katakana (12448 - 12543)
  // Hiragana (12352 - 12447)

static int isKanaChar(gunichar c) {
  if ((c >= 12448) && (c <= 12543)) return TRUE; // Katakana 
  if ((c >= 12352) && (c <= 12447)) return TRUE; // Hiragana
  return FALSE;
}
/*
static int isKatakanaChar(gunichar c) {
  if ((c >= 12448) && (c <= 12543)) return TRUE; // Katakana 
  return FALSE;
}
static int isHiraganaChar(gunichar c) {
  if ((c >= 12352) && (c <= 12447)) return TRUE; // Hiragana
  return FALSE;
}
*/
static int isKanjiChar(gunichar c) {
  if ((c >= 19968) && (c <= 40959)) return TRUE; // Kanji
  return FALSE;
}

void print_dicname() {
  gtk_text_buffer_insert(GTK_TEXT_BUFFER(text_results_buffer), &iter, _("Matches in "), -1);
  gtk_text_buffer_insert_with_tags_by_name(GTK_TEXT_BUFFER(text_results_buffer), &iter,  
					   conf.dicfiles[conf.selected_dic], -1, "brown_foreground", NULL);
  gtk_text_buffer_insert(GTK_TEXT_BUFFER(text_results_buffer), &iter, ":\n", -1);
  dicname_printed = TRUE;
}

static void search_dict(unsigned char *srchstrg) {
  int srchresp, roff, rlen;
  unsigned char repstr[1024];
  uint32_t respos, oldrespos;
  int printit;
  int match_criteria = EXACT_MATCH;
  //int verbdeinf;
  //int vi,vok;
  //char temp[11], vlast[11];
  gchar *currchar;
  int match_type = ANY_MATCH;
  int gjit_search = SRCH_START;
  //gunichar tmpunichar;
  //int i = 0;
  //int contains_kanji, contains_hiragana;

  while (isspace(srchstrg[0])) srchstrg++; // remove leading and trailing spaces
  while (isspace(srchstrg[strlen(srchstrg)-1]) != 0) 
    srchstrg[strlen(srchstrg)-1] = 0;
  if (strlen(srchstrg) == 0) return;
  gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(combo_entry)->entry), srchstrg);
  
  //Detect Japanese
  engsrch = TRUE;
  jpsrch = FALSE;
  currchar = srchstrg;
  do { 
    if (g_unichar_iswide(g_utf8_get_char(currchar)) == TRUE) {
      engsrch = FALSE;
      jpsrch = TRUE;
      break;
    }
  }
  while ((currchar = g_utf8_find_next_char(currchar, srchstrg + strlen(srchstrg))) != NULL);
  
  // Verb deinfelction
  if (GTK_TOGGLE_BUTTON (checkb_verb)->active) print_verb_inflections(srchstrg);
  

  if (jpsrch == TRUE) {
    if (GTK_TOGGLE_BUTTON (radiob_jpexact)->active) match_criteria = EXACT_MATCH;
    if (GTK_TOGGLE_BUTTON (radiob_startw)->active) match_criteria = START_WITH_MATCH;
    if (GTK_TOGGLE_BUTTON (radiob_endw)->active) match_criteria = END_WITH_MATCH;
    if (GTK_TOGGLE_BUTTON (radiob_any)->active) match_criteria = ANY_MATCH;
  }
  else {
    if (GTK_TOGGLE_BUTTON (radiob_engexact)->active) match_criteria = EXACT_MATCH;
    if (GTK_TOGGLE_BUTTON (radiob_words)->active) match_criteria = WORD_MATCH;
    if (GTK_TOGGLE_BUTTON (radiob_partial)->active) match_criteria = ANY_MATCH;
  }


  oldrespos = srchpos = 0;
  
  do { // search loop
    oldrespos = respos;
    
    search4string (gjit_search, conf.selected_dic , srchpos, strlen(srchstrg), srchstrg, 
	       &srchresp, &respos, &roff, &rlen, repstr);
    //printf("respos:%d, oldrespos:%d, roffset:%d, rlen:%d\nrepstr:%s\n", respos, oldrespos, roff, rlen, repstr);
    if (srchresp != SRCH_OK)  {
      return;   //No more matches
    }
    if (gjit_search == SRCH_START) {
      srchpos = respos;
      gjit_search = SRCH_CONT;
    }
    srchpos++;
    if (oldrespos == respos) continue;

    // Check match type and search options
    printit = FALSE;
    if (jpsrch) {
      match_type = get_jp_match_type(repstr, srchstrg, roff);
      switch (match_criteria) {
      case EXACT_MATCH : 
	if (match_type == EXACT_MATCH) printit = TRUE;
	break;
      case START_WITH_MATCH:
	if ((match_type == START_WITH_MATCH) || (match_type == EXACT_MATCH)) printit = TRUE;
	break;
      case END_WITH_MATCH:
	if ((match_type == END_WITH_MATCH) || (match_type == EXACT_MATCH)) printit = TRUE;
	break;
      case ANY_MATCH:
	printit = TRUE;
	break;
      }
    }
    else { //Non-japanese search
      switch (match_criteria) {
      case EXACT_MATCH:
	if ((repstr[roff-1] == '/') && (repstr[roff+strlen(srchstrg)] == '/')) printit = TRUE;
	break;
      case WORD_MATCH:
	if ((g_unichar_isalpha(g_utf8_get_char(repstr + roff + strlen(srchstrg))) == FALSE)  &&
	    (g_unichar_isalpha(g_utf8_get_char(repstr + roff - 1)) == FALSE)) {
	  /*
	  printf("---------\n");
	  printf("%s", repstr);
	  if (g_unichar_isalpha(g_utf8_get_char(repstr + roff - 1)) == FALSE)  printf("beg:%s", repstr + roff - 1);
	  if (g_unichar_isalpha(g_utf8_get_char(repstr + roff + strlen(srchstrg))) == FALSE)  
	    printf("end:%s", repstr + roff + strlen(srchstrg));
	  printf("---------\n");
	  */
	  printit = TRUE;
	}
	break;
      case ANY_MATCH:
	printit = TRUE;
	break;
      }
    }
  
    if (printit) {
      /*
      printf("offset: %d: ", roff);
      printf("jptype: %d\n", match_type);
      printf("criteria: %d\n", match_criteria);
      */
      if ((dicname_printed == FALSE) && (GTK_TOGGLE_BUTTON(radiob_searchall)->active)) {
	print_dicname();
      }
      print_result(repstr, roff, srchstrg);
      get_kanji_and_reading(repstr);
      //printf("%s|%s\n", line_kanji, line_reading);
      word_matches++;
    }
    if ((conf.searchlimit_enabled == TRUE) && (word_matches >= MAX_MATCHES)) {
      gtk_text_buffer_insert_with_tags_by_name(GTK_TEXT_BUFFER(text_results_buffer), &iter,  
					       _("Results truncated"), -1, "red_foreground", NULL);
      return;
    }
  } while (srchresp == SRCH_OK);
}

int lower_search_option() {
  if (!(GTK_TOGGLE_BUTTON(checkb_autoadjust)->active))
    return FALSE;
  if (jpsrch) { //Japanese srting
    if (GTK_TOGGLE_BUTTON(radiob_any)->active)
      return FALSE;
    if (GTK_TOGGLE_BUTTON(radiob_jpexact)->active) {
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radiob_startw), TRUE);
      return TRUE;
    }
    if (GTK_TOGGLE_BUTTON(radiob_startw)->active) {
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radiob_endw), TRUE);
      return TRUE;
    }
    if (GTK_TOGGLE_BUTTON(radiob_endw)->active) {
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radiob_any), TRUE);
      return TRUE;
    }
  }
  else if (engsrch) { //English
    if (GTK_TOGGLE_BUTTON(radiob_partial)->active)
      return FALSE;
    if (GTK_TOGGLE_BUTTON(radiob_engexact)->active) {
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radiob_words), TRUE);
      return TRUE;
    }
    if (GTK_TOGGLE_BUTTON(radiob_words)->active) {
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radiob_partial), TRUE);
      return TRUE;
    }
  }
  return FALSE;
}

void gjiten_search(unsigned char *srchstrg) {
  int Dicnum_save;
  gchar appbarmsg[50];
  int truncated;

  Dicnum_save = conf.selected_dic;
  word_matches = 0;

  if (conf.dicts_ok == FALSE) {
    sprintf(appbarmsg,_("Dicfile error! Set your preferences first."));
    gnome_appbar_set_status(GNOME_APPBAR(appbar_mainwin),appbarmsg);
    return;
  } 
  
  if (conf.numofdics == 0) {
    sprintf(appbarmsg,_("No dicfiles specified! Set your preferences first."));
    gnome_appbar_set_status(GNOME_APPBAR(appbar_mainwin),appbarmsg);
    return;
  } 

  truncated = 0;
  while (TRUE) {
    if (GTK_TOGGLE_BUTTON(radiob_searchall)->active) {
      for (conf.selected_dic = 0; conf.selected_dic < conf.numofdics; conf.selected_dic++) {
	dicname_printed = FALSE;
	search_dict(srchstrg);
	if ((conf.searchlimit_enabled == TRUE) && (word_matches >= MAX_MATCHES))
	  break;
      }
    }
    else search_dict(srchstrg);
    if ((conf.searchlimit_enabled == TRUE) && (word_matches >= MAX_MATCHES))
	  truncated = 1;
    if (word_matches > 0) break;  // No need to search anymore
    if (lower_search_option() == FALSE) break;
  }

  conf.selected_dic = Dicnum_save;
  if (word_matches) {
    if (truncated)
      sprintf(appbarmsg,_("Matches found (truncated): %d"),word_matches);
    else
      sprintf(appbarmsg,_("Matches found: %d"),word_matches);
    gnome_appbar_set_status(GNOME_APPBAR(appbar_mainwin),appbarmsg);
  }
  else gnome_appbar_set_status(GNOME_APPBAR(appbar_mainwin), _("No match found!"));
}



int get_word(char *src, char *dest, int pos) { /*0 if no more words in src, else new pos*/

  int k,j;
  
  k = pos;
  if (src[k] == ' ')  k++;
  if ( strlen(src)-1 <= k) return(0);
  
  j = 0;
  if (src[k] == '{') 
    while (src[k] != '}') {
      dest[j] = src[k];
      j++;
      k++;
    }
  else while (src[k] != ' ') {
    dest[j] = src[k];
    j++;
    k++;
  }
  dest[j] = 0;
  return(k);
}


int init_settings() {
  if (conf.version) { // if gjiten was run before
    if (strcmp (conf.version, VERSION)) {
      if (dialog_preferences == NULL) {
	sprintf(gnome_dialog_msg, 
		_("Config file has a different version.\n Please update your Preferences!"));
      }
    }
  }
  else {
    if (dialog_preferences == NULL) {
      sprintf(gnome_dialog_msg, _("Welcome to Gjiten.\n Please set the Preferences!"));
    }
    return 1; /*no cfg*/
  }

  for (i = 0; i <= 50; i++) {
    if (conf.history[i] == NULL) break;
    if (g_utf8_validate(conf.history[i], -1, NULL) == TRUE)
      combo_entry_glist = g_list_append(combo_entry_glist, g_strdup(conf.history[i]));
    // printf("Read: %s: %s\n", historystr, tmpptr);
  } 
  

  Verbinit();
  if (conf.usekanjidic) load_radkfile();

  return 0;
}


void clear_entry_box(gpointer entrybox) {
  gtk_entry_set_text(GTK_ENTRY(entrybox), "" );
}


void on_text_entered() {
  static gchar *new_entry_text = NULL;

  new_entry_text = g_strdup(gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(combo_entry)->entry)));
  if (g_utf8_validate(new_entry_text, -1, NULL) == FALSE) {
    g_error("INVALID UTF8 STRIG!\n"); //FIXME
    g_free(new_entry_text);
    return;
  }
  if (strlen(new_entry_text) == 0) return;
  if (append_to_history == TRUE) {
    if (current_glist_word != NULL) {
      if (!(strcmp((char*) g_list_find(combo_entry_glist, current_glist_word)->data, new_entry_text) == 0)) {
	current_glist_word = new_entry_text;
	combo_entry_glist = g_list_prepend(combo_entry_glist, new_entry_text);
      }
    }
    else {
      current_glist_word = new_entry_text;
      combo_entry_glist = g_list_prepend(combo_entry_glist, new_entry_text);
    }
  }

  if (g_list_next(g_list_find(combo_entry_glist, current_glist_word)) != NULL) 
    gtk_widget_set_sensitive(button_back, TRUE);
  else gtk_widget_set_sensitive(button_back, FALSE);

  if (g_list_previous(g_list_find(combo_entry_glist, current_glist_word)) != NULL) 
    gtk_widget_set_sensitive(button_forward, TRUE);
  else gtk_widget_set_sensitive(button_forward, FALSE);

  gtk_text_buffer_set_text (GTK_TEXT_BUFFER(text_results_buffer), "", 0);
  gtk_text_buffer_get_start_iter(text_results_buffer, &iter);

  gnome_appbar_set_status(GNOME_APPBAR(appbar_mainwin), _("Searching..."));

  gjiten_search(new_entry_text);
  
  if (append_to_history == FALSE)
    g_free(new_entry_text);
  else 
    gtk_combo_set_popdown_strings(GTK_COMBO(combo_entry), combo_entry_glist);
}

void on_forward_clicked() { 
  append_to_history = FALSE;
  current_glist_word = (gchar*) g_list_previous(g_list_find(combo_entry_glist, current_glist_word))->data;
  gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(combo_entry)->entry), current_glist_word);
  on_text_entered();  
  append_to_history = TRUE;
}

void on_back_clicked() {
  append_to_history = FALSE;
  current_glist_word = (gchar*) g_list_next(g_list_find(combo_entry_glist, current_glist_word))->data;
  gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(combo_entry)->entry), current_glist_word);
  on_text_entered();
  append_to_history = TRUE;
}

void on_dicselection_clicked(int dicnum) {
  conf.selected_dic = dicnum;
}

void display_manual(GtkWidget *widget, void *data) {
  GtkWidget *window = data;
  GError *err;
  
  err = NULL;  
  gnome_help_display("gjiten.xml", NULL, &err);
  
  if (err != NULL) {
    GtkWidget *dialog;
    dialog = gtk_message_dialog_new(GTK_WINDOW(window),
				     GTK_DIALOG_DESTROY_WITH_PARENT,       
				     GTK_MESSAGE_ERROR,
				     GTK_BUTTONS_CLOSE,
				     _("Could not display help: %s"),
				     err->message);
    
    g_signal_connect(G_OBJECT(dialog), "response",
		      G_CALLBACK(gtk_widget_destroy),
		      NULL);
    
    gtk_window_set_resizable(GTK_WINDOW(dialog), FALSE);
    
    gtk_widget_show(dialog);
    
    g_error_free(err);
  }
}


void create_about() {
  const gchar *authors[] = { "Botond Botyanszki <boti@rocketmail.com>", NULL };
  const gchar *documenters[] = { NULL };
  const gchar *translator = _("translated_by");
  static GtkWidget *about = NULL;
  GdkPixbuf *pixbuf = NULL;

  if (about != NULL) {
    gdk_window_show(about->window);
    gdk_window_raise(about->window);
    return;
  }

  if (pixbuf != NULL) {
    GdkPixbuf* temp_pixbuf = NULL;
    
    temp_pixbuf = gdk_pixbuf_scale_simple(pixbuf, 
					   gdk_pixbuf_get_width(pixbuf) / 2,
					   gdk_pixbuf_get_height(pixbuf) / 2,
					   GDK_INTERP_HYPER);
    g_object_unref(pixbuf);
    pixbuf = temp_pixbuf;
  }

  pixbuf = gdk_pixbuf_new_from_file(PIXMAPDIR"gjiten/gjiten-logo.png", NULL);

  if (strcmp(translator, "translated_by") == 0) translator = NULL;

  /*
    _("Released under the terms of the GNU GPL.\n"
    "Check out http://gjiten.sourceforge.net for updates"), 
  */
  about = gnome_about_new("gjiten", VERSION, "Copyright \xc2\xa9 1999-2003 Botond Botyanszki",
			  _("gjiten is a Japanese dictionary for Gnome"),
			  (const char **)authors,
			  (const char **)documenters,
			  (const char *)translator,
			  pixbuf);

  gtk_window_set_destroy_with_parent(GTK_WINDOW(about), TRUE);
  if (pixbuf != NULL)  g_object_unref (pixbuf);

  g_signal_connect(G_OBJECT(about), "destroy", G_CALLBACK(gtk_widget_destroyed), &about);
  gtk_widget_show(about);

}

void checkb_searchlimit_toggled() {
  int state = GTK_TOGGLE_BUTTON(checkb_searchlimit)->active;
  if (spinb_searchlimit != NULL) gtk_widget_set_sensitive(spinb_searchlimit, state);
  conf.searchlimit_enabled = state;
  if (conf.maxwordmatches == 0) conf.searchlimit_enabled = FALSE;
}

void shade_gjiten_widgets() {
  if ((menu_selectdic != NULL) && (radiob_searchdic != NULL)) 
    gtk_widget_set_sensitive(menu_selectdic, GTK_TOGGLE_BUTTON(radiob_searchdic)->active);

  if (checkb_autoadjust != NULL) conf.autoadjust_enabled = (GTK_TOGGLE_BUTTON(checkb_autoadjust)->active);
  if (checkb_verb != NULL) conf.deinflection_enabled = (GTK_TOGGLE_BUTTON(checkb_verb)->active);  
}

void get_searchlimit() {
  conf.maxwordmatches = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spinb_searchlimit));
  if (conf.maxwordmatches == 0) conf.searchlimit_enabled = FALSE;
}

void create_window_gjiten() {
  GtkWidget *vbox_main;
  GtkWidget *dock_main;
  GtkWidget *toolbar;
  GtkWidget *button_exit;
  //  GtkWidget *button_pref;
  GtkWidget *button_kanjipad;
  GtkWidget *button_kanjidic;
  GtkWidget *button_clear;
  GtkWidget *button_srch;
  //  GtkWidget *button_copy;
  //  GtkWidget *button_paste;
  GtkWidget *hbox_options;
  GtkWidget *frame_japopt;
  GtkWidget *vbox_japopt;
  GSList *vbox_japopt_group = NULL;
  GSList *dicssearch_group = NULL;
  GtkWidget *frame_engopt;
  GtkWidget *vbox_engopt;
  GSList *vbox_engopt_group = NULL;
  GtkWidget *table_gopt;
  GtkWidget *frame_gopt;
  GtkWidget *menu_dictfiles_item;
  GtkWidget *hbox_searchlimit;
  GtkWidget *hbox_entry;
  GtkWidget *label_enter;
  GtkWidget *button_search;
  GtkWidget *frame_results;
  GtkWidget *vbox_results;
  GtkWidget *scrolledwin_results;
  GtkObject *spinb_searchlimit_adj;
  GtkWidget *tmpimage;
  GtkWidget *dicselection_menu;

  spinb_searchlimit = NULL;
  radiob_searchdic = NULL;
  checkb_autoadjust = NULL;
  checkb_verb = NULL;

  window_gjiten = gnome_app_new("gjiten", "Gjiten - WordDic");
  GTK_WIDGET_SET_FLAGS(window_gjiten, GTK_CAN_DEFAULT);
  g_signal_connect(GTK_OBJECT(window_gjiten), "destroy", GTK_SIGNAL_FUNC(gjiten_exit), NULL);
  gtk_window_set_default_size(GTK_WINDOW(window_gjiten), 500, 500);

  dock_main = GNOME_APP(window_gjiten)->dock;
  gtk_widget_show(dock_main);

  if (conf.menubar) gnome_app_create_menus(GNOME_APP(window_gjiten), menubar_uiinfo);

  if (conf.toolbar) {
    toolbar = gtk_toolbar_new();
    gtk_widget_show(toolbar);

    //gnome_app_create_toolbar(GNOME_APP(window_gjiten), toolbar_uuinfo);
    gnome_app_set_toolbar(GNOME_APP(window_gjiten), GTK_TOOLBAR(toolbar));
    
    button_exit = gtk_toolbar_insert_stock(GTK_TOOLBAR(toolbar), GTK_STOCK_CLOSE,
					   _("Close"), "Close", 
					   gjiten_exit, NULL, -1);

    button_back = gtk_toolbar_insert_stock(GTK_TOOLBAR(toolbar), GTK_STOCK_GO_BACK,
					   _("Back"), "Back", 
					   on_back_clicked, NULL, -1);
    gtk_widget_set_sensitive(button_back, FALSE);

    button_forward = gtk_toolbar_insert_stock(GTK_TOOLBAR(toolbar), GTK_STOCK_GO_FORWARD,
					      _("Forward"), "Forward", 
					      on_forward_clicked, NULL, -1);
    gtk_widget_set_sensitive(button_forward, FALSE);


    tmpimage = gtk_image_new_from_file(PIXMAPDIR"kanjidic.png");
    button_kanjidic = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar), _("KanjiDic"),
					      _("KanjiDic"), "KanjiDic", tmpimage,
					      GTK_SIGNAL_FUNC(create_window_kanjidic), NULL);
    gtk_widget_set_sensitive(button_kanjidic, conf.usekanjidic);

    tmpimage = gtk_image_new_from_file(PIXMAPDIR"kanjipad.png");
    button_kanjipad = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar), _("KanjiPad"),
					      _("KanjiPad"), "KanjiPad", tmpimage,
					      GTK_SIGNAL_FUNC(start_kanjipad), NULL);
    gtk_widget_set_sensitive(button_kanjipad, conf.usekanjipad);

    button_srch = gtk_toolbar_insert_stock(GTK_TOOLBAR(toolbar), GTK_STOCK_FIND,
					       _("Search"), "Search", 
					       on_text_entered, NULL, -1);


    /*
    button_srch = gtk_toolbar_insert_item(GTK_TOOLBAR(toolbar), _("Search"), "Search", "Search", 
                                             GtkWidget *icon,
					  on_text_entered, NULL, -1);
    */
    /*

    tmp_toolbar_icon = gnome_stock_pixmap_widget(window_gjiten, GNOME_STOCK_PIXMAP_COPY);
    button_copy = gtk_toolbar_append_element(GTK_TOOLBAR(toolbar),
					      GTK_TOOLBAR_CHILD_BUTTON,  NULL, _("Copy"),
					      NULL, NULL, tmp_toolbar_icon, NULL, NULL);
    gtk_widget_show(button_copy);
    g_signal_connect(GTK_OBJECT(button_copy), "clicked",GTK_SIGNAL_FUNC(gjiten_copy), NULL);
  
    tmp_toolbar_icon = gnome_stock_pixmap_widget(window_gjiten, GNOME_STOCK_PIXMAP_PASTE);
    button_paste = gtk_toolbar_append_element(GTK_TOOLBAR(toolbar),
					       GTK_TOOLBAR_CHILD_BUTTON, NULL,_("Paste"),
					       NULL, NULL, tmp_toolbar_icon, NULL, NULL);
    gtk_widget_show(button_paste);
    g_signal_connect(GTK_OBJECT(button_paste), "clicked",GTK_SIGNAL_FUNC(gjiten_paste), NULL);

    */   
    gtk_toolbar_set_style(GTK_TOOLBAR(toolbar), GTK_TOOLBAR_BOTH);

  }    
    


  vbox_main = gtk_vbox_new(FALSE, 0);
  gtk_widget_show(vbox_main);
  gnome_app_set_contents(GNOME_APP(window_gjiten), vbox_main);

  hbox_options = gtk_hbox_new(FALSE, 0);
  gtk_widget_show(hbox_options);
  gtk_box_pack_start(GTK_BOX(vbox_main), hbox_options, FALSE, TRUE, 0);

  frame_japopt = gtk_frame_new(_("Japanese Search Options: "));
  gtk_widget_show(frame_japopt);
  gtk_box_pack_start(GTK_BOX(hbox_options), frame_japopt, TRUE, TRUE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(frame_japopt), 5);

  vbox_japopt = gtk_vbox_new(FALSE, 0);
  gtk_widget_show(vbox_japopt);
  gtk_container_add(GTK_CONTAINER(frame_japopt), vbox_japopt);

  radiob_jpexact = gtk_radio_button_new_with_label(vbox_japopt_group, _("Exact Matches"));
  vbox_japopt_group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(radiob_jpexact));
  gtk_widget_show(radiob_jpexact);
  gtk_box_pack_start(GTK_BOX(vbox_japopt), radiob_jpexact, FALSE, FALSE, 0);

  radiob_startw = gtk_radio_button_new_with_label(vbox_japopt_group, _("Start With Expression"));
  vbox_japopt_group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(radiob_startw));
  gtk_widget_show(radiob_startw);
  gtk_box_pack_start(GTK_BOX(vbox_japopt), radiob_startw, FALSE, FALSE, 0);

  radiob_endw = gtk_radio_button_new_with_label(vbox_japopt_group, _("End With Expression"));
  vbox_japopt_group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(radiob_endw));
  gtk_widget_show(radiob_endw);
  gtk_box_pack_start(GTK_BOX(vbox_japopt), radiob_endw, FALSE, FALSE, 0);

  radiob_any = gtk_radio_button_new_with_label(vbox_japopt_group, _("Any Matches"));
  vbox_japopt_group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(radiob_any));
  gtk_widget_show(radiob_any);
  gtk_box_pack_start(GTK_BOX(vbox_japopt), radiob_any, FALSE, FALSE, 0);

  checkb_verb = gtk_check_button_new_with_label(_("Verb Deinflection"));
  gtk_widget_show(checkb_verb);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkb_verb), TRUE);
  gtk_box_pack_start(GTK_BOX(vbox_japopt), checkb_verb, FALSE, FALSE, 0);
  g_signal_connect(GTK_OBJECT(checkb_verb), "toggled", 
		      GTK_SIGNAL_FUNC(shade_gjiten_widgets), NULL);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkb_verb), conf.deinflection_enabled);
  
  frame_engopt = gtk_frame_new(_("English Search Options: "));
  gtk_widget_show(frame_engopt);
  gtk_box_pack_start(GTK_BOX(hbox_options), frame_engopt, TRUE, TRUE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(frame_engopt), 5);

  vbox_engopt = gtk_vbox_new(FALSE, 0);
  gtk_widget_show(vbox_engopt);
  gtk_container_add(GTK_CONTAINER(frame_engopt), vbox_engopt);

  radiob_engexact = gtk_radio_button_new_with_label(vbox_engopt_group, 
						     _("Whole Expressions"));
  vbox_engopt_group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(radiob_engexact));
  gtk_widget_show(radiob_engexact);
  gtk_box_pack_start(GTK_BOX(vbox_engopt), radiob_engexact, FALSE, FALSE, 0);

  radiob_words = gtk_radio_button_new_with_label(vbox_engopt_group, _("Whole Words"));
  vbox_engopt_group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(radiob_words));
  gtk_widget_show(radiob_words);
  gtk_box_pack_start(GTK_BOX(vbox_engopt), radiob_words, FALSE, FALSE, 0);

  radiob_partial = gtk_radio_button_new_with_label(vbox_engopt_group, _("Any Matches"));
  vbox_engopt_group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(radiob_partial));
  gtk_widget_show(radiob_partial);
  gtk_box_pack_start(GTK_BOX(vbox_engopt), radiob_partial, FALSE, FALSE, 0);

  frame_gopt = gtk_frame_new(_("General Options: "));
  gtk_widget_show(frame_gopt);
  gtk_box_pack_start(GTK_BOX(hbox_options), frame_gopt, TRUE, TRUE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(frame_gopt), 5);

  table_gopt = gtk_table_new(3, 2, FALSE);
  gtk_widget_show(table_gopt);
  gtk_container_add(GTK_CONTAINER(frame_gopt), table_gopt);

  radiob_searchdic = gtk_radio_button_new_with_label(dicssearch_group, _("Search Dic:"));
  dicssearch_group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(radiob_searchdic));
  gtk_widget_show(radiob_searchdic);
  gtk_table_attach(GTK_TABLE(table_gopt), radiob_searchdic, 0, 1, 0, 1,
                   (GtkAttachOptions) (GTK_EXPAND | GTK_FILL), 
		   (GtkAttachOptions)(0), 0, 0);
  g_signal_connect_swapped(GTK_OBJECT(radiob_searchdic), "clicked", 
			   GTK_SIGNAL_FUNC(shade_gjiten_widgets), NULL);

  // DICTFILE SELECTION MENU
  menu_selectdic = gtk_menu_new();

  for (i = 0; i < conf.numofdics ; i++) {
    menu_dictfiles_item = gtk_menu_item_new_with_label(conf.dicfiles[i]);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu_selectdic), menu_dictfiles_item);
    g_signal_connect_swapped(GTK_OBJECT(menu_dictfiles_item), "activate", 
			     GTK_SIGNAL_FUNC(on_dicselection_clicked), (gpointer) i);
    gtk_widget_show(menu_dictfiles_item);
  }

  dicselection_menu = gtk_option_menu_new();
  gtk_widget_show(dicselection_menu);
  gtk_option_menu_set_menu(GTK_OPTION_MENU(dicselection_menu), menu_selectdic);

  gtk_table_attach(GTK_TABLE(table_gopt), dicselection_menu, 1, 2, 0, 1,
                   (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
		   (GtkAttachOptions)(0), 0, 0);
  //

  radiob_searchall = gtk_radio_button_new_with_label(dicssearch_group,
						      _("Search All Dictionaries"));
  gtk_widget_show(radiob_searchall);
  gtk_table_attach(GTK_TABLE(table_gopt), radiob_searchall, 0, 2, 1, 2,
                   (GtkAttachOptions)(GTK_EXPAND | GTK_FILL), 
		   (GtkAttachOptions)(0), 0, 0);
  g_signal_connect_swapped(GTK_OBJECT(radiob_searchall), "clicked", 
			     GTK_SIGNAL_FUNC(shade_gjiten_widgets), NULL);


  checkb_autoadjust = gtk_check_button_new_with_label(_("Auto Adjust Options"));
  gtk_widget_show(checkb_autoadjust);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkb_autoadjust), TRUE);
  gtk_table_attach(GTK_TABLE(table_gopt), checkb_autoadjust, 0, 2, 2, 3,
                   (GtkAttachOptions)(GTK_EXPAND | GTK_FILL), 
		   (GtkAttachOptions)(0), 0, 0);
  g_signal_connect(GTK_OBJECT(checkb_autoadjust), "toggled", 
		      GTK_SIGNAL_FUNC(shade_gjiten_widgets), NULL);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkb_autoadjust), conf.autoadjust_enabled);

  
  hbox_searchlimit = gtk_hbox_new(FALSE, 0);
  gtk_widget_show(hbox_searchlimit);
  gtk_table_attach(GTK_TABLE(table_gopt), hbox_searchlimit, 0, 2, 3, 4,
                   (GtkAttachOptions)(GTK_FILL), 
		   (GtkAttachOptions)(0), 0, 0);
  checkb_searchlimit = gtk_check_button_new_with_label(_("Limit Results:"));
  gtk_widget_show(checkb_searchlimit);
  gtk_box_pack_start(GTK_BOX(hbox_searchlimit), checkb_searchlimit, FALSE, FALSE, 0);
  g_signal_connect(GTK_OBJECT(checkb_searchlimit), "toggled", 
		      GTK_SIGNAL_FUNC(checkb_searchlimit_toggled), NULL);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkb_searchlimit), conf.searchlimit_enabled);

  spinb_searchlimit_adj = gtk_adjustment_new(conf.maxwordmatches, 1, G_MAXFLOAT, 1, 2, 2);
  spinb_searchlimit = gtk_spin_button_new(GTK_ADJUSTMENT(spinb_searchlimit_adj), 1, 0);
  gtk_widget_show(spinb_searchlimit);
  gtk_box_pack_start(GTK_BOX(hbox_searchlimit), spinb_searchlimit, FALSE, FALSE, 0);
  gtk_widget_set_sensitive(spinb_searchlimit,(GTK_TOGGLE_BUTTON(checkb_searchlimit)->active));
  g_signal_connect(GTK_OBJECT(spinb_searchlimit_adj), "value_changed", 
		      GTK_SIGNAL_FUNC(get_searchlimit), NULL);
  

  hbox_entry = gtk_hbox_new(FALSE, 0);   
  gtk_widget_show(hbox_entry);
  gtk_box_pack_start(GTK_BOX(vbox_main), hbox_entry, FALSE, TRUE, 14);
  gtk_container_set_border_width(GTK_CONTAINER(hbox_entry), 3);

  label_enter = gtk_label_new(_("Enter expression :"));
  gtk_widget_show(label_enter);
  gtk_box_pack_start(GTK_BOX(hbox_entry), label_enter, FALSE, TRUE, 5);
  gtk_label_set_justify(GTK_LABEL(label_enter), GTK_JUSTIFY_RIGHT);
  gtk_misc_set_alignment(GTK_MISC(label_enter), 1, 0.5);
  gtk_misc_set_padding(GTK_MISC(label_enter), 7, 0);

  combo_entry = gtk_combo_new();
  gtk_widget_show(combo_entry);
  gtk_box_pack_start(GTK_BOX(hbox_entry), combo_entry, TRUE, TRUE, 0);
  //gtk_widget_set_size_request(combo_entry, -2, 25);
  g_signal_connect(GTK_OBJECT(GTK_COMBO(combo_entry)->entry), 
		      "activate", GTK_SIGNAL_FUNC(on_text_entered), NULL);
  gtk_combo_disable_activate(GTK_COMBO(combo_entry));
  gtk_combo_set_case_sensitive(GTK_COMBO(combo_entry), TRUE);
  if (combo_entry_glist != NULL) {
    gtk_combo_set_popdown_strings(GTK_COMBO(combo_entry), combo_entry_glist);
    gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(combo_entry)->entry), "");
  }
  GTK_WIDGET_SET_FLAGS(GTK_COMBO(combo_entry)->entry, GTK_CAN_DEFAULT);
  gtk_widget_grab_focus(GTK_COMBO(combo_entry)->entry);
  gtk_widget_grab_default(GTK_COMBO(combo_entry)->entry);
  

  button_search = gtk_button_new_with_label(_("Search"));
  gtk_widget_show(button_search);
  gtk_box_pack_start(GTK_BOX(hbox_entry), button_search, FALSE, FALSE, 7);
  //gtk_widget_set_size_request(button_search, 81, -2);
  g_signal_connect(GTK_OBJECT(button_search), "clicked",GTK_SIGNAL_FUNC(on_text_entered), NULL);

  button_clear = gtk_button_new_with_label(_("Clear"));
  gtk_widget_show(button_clear);
  gtk_box_pack_start(GTK_BOX(hbox_entry), button_clear, FALSE, FALSE, 0);
  //gtk_widget_set_size_request(button_clear, 47, -2);
  g_signal_connect_swapped(GTK_OBJECT(button_clear), "clicked", 
			   GTK_SIGNAL_FUNC(clear_entry_box), 
			   GTK_OBJECT(GTK_COMBO(combo_entry)->entry));

  frame_results = gtk_frame_new(_("Search results :"));
  gtk_widget_show(frame_results);
  gtk_box_pack_start(GTK_BOX(vbox_main), frame_results, TRUE, TRUE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(frame_results), 5);
  gtk_frame_set_label_align(GTK_FRAME(frame_results), 0.03, 0.5);

  vbox_results = gtk_vbox_new(FALSE, 0);
  gtk_widget_show(vbox_results);
  gtk_container_add(GTK_CONTAINER(frame_results), vbox_results);

  text_results_view = gtk_text_view_new();
  text_results_buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(text_results_view));
  gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(text_results_view), GTK_WRAP_WORD);

  gtk_widget_show(text_results_view);
  //FIXME  gtk_text_set_editable(GTK_TEXT_BUFFER(text_results_view), FALSE);
  
  //set up fonts and tags
  gtk_text_buffer_create_tag(text_results_buffer, "blue_foreground", "foreground", "blue", NULL);  
  gtk_text_buffer_create_tag(text_results_buffer, "red_foreground", "foreground", "red", NULL);  
  gtk_text_buffer_create_tag(text_results_buffer, "brown_foreground", "foreground", "brown", NULL);  
  if (conf.bigfont == NULL) {
    gtk_text_buffer_create_tag(text_results_buffer, "bigfont", "size", 20 * PANGO_SCALE, NULL);
  }
  else {
    gtk_text_buffer_create_tag(text_results_buffer, "bigfont", "font", conf.bigfont, NULL);
  }
  if ((conf.normalfont != NULL) && conf.normalfont_desc != NULL) {
    gtk_widget_modify_font(text_results_view, conf.normalfont_desc);
    gtk_widget_modify_font(GTK_COMBO(combo_entry)->entry, conf.normalfont_desc);
  }
  
  scrolledwin_results = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrolledwin_results), GTK_SHADOW_IN);

  gtk_container_add(GTK_CONTAINER(scrolledwin_results), text_results_view);
  gtk_box_pack_start(GTK_BOX(vbox_results), scrolledwin_results, TRUE, TRUE, 0);
  gtk_widget_show(scrolledwin_results);

  appbar_mainwin = gnome_appbar_new(TRUE, TRUE, GNOME_PREFERENCES_NEVER);
  gtk_widget_show(appbar_mainwin);
  gtk_box_pack_end(GTK_BOX(vbox_results), appbar_mainwin, FALSE, FALSE, 0);
 
  gtk_widget_show(window_gjiten);

  show_errors();

}


/*********************** MAIN ***********************************/


int main (int argc, char **argv) {
  char *icon_path = PIXMAPDIR"/jiten.png";
  GError *error;

  set_strginfo();

  gnome_dialog_msg = (gchar *)g_malloc(300);
  if (gnome_dialog_msg == NULL) exit_critical("Couldn't allocate memory\n");
  gnome_dialog_msg[0] = 0;

  if (gconf_init(argc, argv, &error) == FALSE) {
    g_error("GConf init failed: %s\n", error->message);
    exit(-1);
  }
  
  g_type_init();
  gconf_client = gconf_client_get_default();
  if (load_settings() == 1) conf.dicts_ok = FALSE;
  if (conf.gdk_use_xft == TRUE) putenv("GDK_USE_XFT=1");
  else putenv("GDK_USE_XFT=0");
  if (conf.force_ja_JP == TRUE) putenv("LC_ALL=ja_JP");
  if (conf.force_language_c == TRUE) putenv("LANGUAGE=C");

  gtk_set_locale();  

#ifdef ENABLE_NLS
  bindtextdomain(PACKAGE, PACKAGE_LOCALE_DIR);
  bind_textdomain_codeset(PACKAGE, "UTF-8");
  textdomain(PACKAGE);
#endif

#ifndef USE_MMAP
  linsrchbuff = (gchar *)g_malloc(LINBUFFSIZE);
  if (linsrchbuff == NULL) exit_critical("Couldn't allocate memory\n");  
#endif

  gnome_program_init("gjiten", VERSION, LIBGNOMEUI_MODULE, argc, argv, 
		     GNOME_PARAM_POPT_TABLE, arg_options, 
		     GNOME_PARAM_HUMAN_READABLE_NAME, _("gjiten"), 
		     GNOME_PARAM_APP_DATADIR, DATADIR,
		     NULL);

  if (! g_file_test (icon_path, G_FILE_TEST_EXISTS)) {
    g_warning ("Could not find %s", icon_path);
  }
  else {
    gnome_window_icon_set_default_from_file(icon_path);
  }

  cmap = gdk_colormap_get_system();
  color1.red = 0;
  color1.green = 0;
  color1.blue = 0xb000;
  color2.red = 0x6000;
  color2.green = 0x9000;
  color2.blue = 0x0100;
  if (!gdk_color_alloc(cmap, &color1)) {
    g_warning("couldn't allocate colour");
  }

  conf.selected_dic = 0; /*first dic in conf diclist*/
  init_settings();
 
  // the following is for clipboard lookup.
  if ((conf.clip_kanji_lookup == TRUE) || (conf.clip_word_lookup == TRUE)) {
    if (conf.clip_word_lookup) {
      create_window_gjiten();
      gjiten_paste();
      on_text_entered();
    }
    else {
      if (conf.clip_kanji_lookup){
	clipboard_text = gtk_clipboard_wait_for_text(gtk_clipboard_get(GDK_SELECTION_PRIMARY));
	//validate
	//FIXME: try to convert EUC-JP to UTF8 if it's non-utf8
	if (g_utf8_validate(clipboard_text, -1, NULL) == FALSE) {
	  sprintf(gnome_dialog_msg, _("Unable to look up kanji: NON-UTF8 string received from clipboard!\n"));
	  show_errors();
	  exit(0);
	}
	else if (isKanjiChar(g_utf8_get_char(clipboard_text)) == FALSE) {
	  sprintf(gnome_dialog_msg , _("Non-kanji string received from clipboard: %s\n"), clipboard_text);
	  show_errors();
	  exit(0);
	}
	else {
	  if (window_kanjidic == NULL) create_window_kanjidic();
	  print_kanjinfo(g_utf8_get_char(clipboard_text));
	  //if (clipboard_text != NULL) gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(combo_entry)->entry), clipboard_text);
	}
      }
    }
  }

  if (argc > 1) {
    if (conf.startkanjidic) {
      create_window_kanjidic();
    }
    else  
      if (conf.word_to_lookup) {
	create_window_gjiten();
	gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(combo_entry)->entry), conf.word_to_lookup);
	on_text_entered();
      }
      else  
	if (conf.kanji_to_lookup != NULL) {
	  if (g_utf8_validate(conf.kanji_to_lookup, -1, NULL) == FALSE) {
	    sprintf(gnome_dialog_msg, _("Unable to look up kanji: NON-UTF8 string received from clipboard!\n"));
	    show_errors();
	    exit(0);
	  }
	  else 
	    if (isKanjiChar(g_utf8_get_char(conf.kanji_to_lookup)) == FALSE) {
	      sprintf(gnome_dialog_msg , _("Non-kanji string received from clipboard: %s\n"), conf.kanji_to_lookup);
	      show_errors();
	      exit(0);
	    }
	    else {
	      if (window_kanjidic == NULL) create_window_kanjidic();
	      print_kanjinfo(g_utf8_get_char(conf.kanji_to_lookup));
	    }
	}
      	else if (!conf.clip_kanji_lookup && !conf.clip_word_lookup) 
	  create_window_gjiten();
  }
  else {
    create_window_gjiten();
  }
  show_errors();
  gtk_main();
  return 0;
}
