/* pref.c
   
   GJITEN : A GTK+/GNOME BASED JAPANESE DICTIONARY
  
   Copyright (C) 1999-2003 Botond Botyanszki <boti at rocketmail dot com>
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published  by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <gnome.h>

#include "gjiten.h"
#include "kanjidic.h"
#include "gjiten-settings.h"

extern int init_settings();
extern int load_settings();
extern void fonts_init();
extern void create_window_gjiten();
extern void create_window_kanjidic();
extern void show_errors();

extern int shutdown_gjiten;
extern gchar *gnome_dialog_msg;
GtkWidget *dict_list_view;
extern gchar *kanjidicstrg[];
GtkWidget *checkb_prefs[KCFGNUM];
GtkWidget *combo_entry_dictpath;
GtkWidget *combo_entry_kanjidic;
GtkWidget *combo_entry_kanjipad;
//GtkWidget *checkb_tooltips;
GtkWidget *checkb_toolbar;
GtkWidget *checkb_menubar;
GtkWidget *checkb_kanjipad;
GtkWidget *checkb_kanjidic;
GtkWidget *checkb_bigwords;
GtkWidget *checkb_bigkanji;
GtkWidget *checkb_use_xft;
GtkWidget *checkb_force_ja_JP;
GtkWidget *checkb_force_language_c;
extern GtkWidget *window_gjiten;
extern GtkWidget *window_kanjidic;
GtkWidget *dialog_preferences = NULL;
GtkWidget *entry_bigfont;
GtkWidget *entry_normalfont;
extern gchar *gnome_dialog_msg;
GtkTreeIter list_iter;
GtkListStore *list_store;
GtkCellRenderer *cell_renderer;
GtkTreeViewColumn *list_column;
GtkTreeModel *model;
GtkWidget *fileselector;

gchar *strginfo[KCFGNUM];
char fontteststrg[] = {'N', 'i', 'h', 'o', 'n', 'g', 'o', ':', 0xe6, 0x97, 0xa5, 
		       0xe6, 0x9c, 0xac, 0xe8, 0xaa, 0x9e};

enum {
  STRING_COLUMN,
  N_COLUMNS
};



void font_set(gpointer ptr, gchar *arg1, GnomeFontPicker *fontpicker) {
  GtkWidget *entry = (GtkWidget *)ptr;
  const gchar *fontname;

  fontname = gnome_font_picker_get_font_name(GNOME_FONT_PICKER(fontpicker));
  gtk_entry_set_text(GTK_ENTRY(entry), fontname);
}

void dicfile_selected(GtkWidget *fsel, GtkFileSelection *fs) {
  gchar *filename, *dictf, *old;

  if (++conf.numofdics >= MAXDICFILES ) {
    sprintf(gnome_dialog_msg, _("Sorry, the limit is 100 dictionaries."));
    show_errors();
    conf.numofdics = MAXDICFILES;
    return;
  }
   
  filename = (gchar *) gtk_file_selection_get_filename(GTK_FILE_SELECTION(fs));
  old = dictf = strtok(filename,"/");
  while (dictf != NULL) {
    old = dictf;
    dictf = strtok(NULL, "/");
  }
  gtk_list_store_append(list_store, &list_iter);
  gtk_list_store_set(list_store, &list_iter, STRING_COLUMN, old, -1);
  gtk_widget_destroy(fileselector);
}

void add_dict() {
  gchar *dictpath = g_strdup(gtk_entry_get_text(GTK_ENTRY(combo_entry_dictpath)));

  fileselector = gtk_file_selection_new(_("Select Dictionary Files"));
  gtk_file_selection_set_filename(GTK_FILE_SELECTION(fileselector), dictpath);
  /*printf("%s\n",conf.dictpath);
   */
  gtk_widget_show(fileselector);
  g_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(fileselector)->ok_button),"clicked",
		     GTK_SIGNAL_FUNC(dicfile_selected), fileselector);
  g_signal_connect_swapped(GTK_OBJECT(GTK_FILE_SELECTION(fileselector)->cancel_button),
			    "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), 
			    GTK_OBJECT(fileselector));
}

gchar *get_selected_row() {
  GtkTreePath *path;
  GtkTreeSelection *selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(dict_list_view));

  if (gtk_tree_selection_get_selected(selection, &model, &list_iter) == TRUE) {
    path = gtk_tree_model_get_path(model, &list_iter);
    return gtk_tree_path_to_string(path);
  }
  else return NULL;
}

/*
void print_selection() {
  printf("selected: %s\n", get_selected_row());
}
*/

void down_dict() {
  gchar lastdic[5];
  GtkTreeIter tmpiter;
  GtkTreePath *tmppath;
  gchar *str_data_up;
  gchar *str_data_down;
  gchar *selected;
  GtkTreeSelection *selection;

  sprintf(lastdic, "%d", conf.numofdics - 1);
  selected = get_selected_row();
  if ((selected != NULL) && (strcmp(lastdic, selected) != 0)) {
    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(dict_list_view));
    tmppath = gtk_tree_path_new_from_string(selected);
    gtk_tree_path_next(tmppath);
    gtk_tree_model_get_iter(model, &tmpiter, tmppath);
    gtk_tree_path_free(tmppath);
    gtk_tree_model_get(model, &tmpiter, STRING_COLUMN, &str_data_up, -1);
    gtk_tree_model_get(model, &list_iter, STRING_COLUMN, &str_data_down, -1);
    //printf("%s, %s\n", str_data_up, str_data_down);
    gtk_list_store_set(list_store, &tmpiter, STRING_COLUMN, str_data_down, -1);
    gtk_list_store_set(list_store, &list_iter, STRING_COLUMN, str_data_up, -1);
    tmppath = gtk_tree_path_new_from_string(selected);
    gtk_tree_path_next(tmppath);
    gtk_tree_selection_select_path(selection, tmppath);
  }
}

void up_dict() {
  GtkTreeIter tmpiter;
  GtkTreePath *tmppath;
  gchar *str_data_up;
  gchar *str_data_down;
  gchar *selected;
  GtkTreeSelection *selection;

  selected = get_selected_row();
  if ((selected != NULL) && (strcmp("0", selected) != 0)) {
    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(dict_list_view));
    tmppath = gtk_tree_path_new_from_string(selected);
    gtk_tree_path_prev(tmppath);
    gtk_tree_model_get_iter(model, &tmpiter, tmppath);
    gtk_tree_path_free(tmppath);
    gtk_tree_model_get(model, &tmpiter, STRING_COLUMN, &str_data_up, -1);
    gtk_tree_model_get(model, &list_iter, STRING_COLUMN, &str_data_down, -1);
    //printf("%s, %s\n", str_data_up, str_data_down);
    gtk_list_store_set(list_store, &tmpiter, STRING_COLUMN, str_data_down, -1);
    gtk_list_store_set(list_store, &list_iter, STRING_COLUMN, str_data_up, -1);
    tmppath = gtk_tree_path_new_from_string(selected);
    gtk_tree_path_prev(tmppath);
    gtk_tree_selection_select_path(selection, tmppath);
  }
}

void del_dict() {
  GtkTreePath *tmppath;
  gchar *selected;
  GtkTreeSelection *selection;

  selected = get_selected_row();
  if (selected != NULL) {
    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(dict_list_view));
    tmppath = gtk_tree_path_new_from_string(selected);
    gtk_list_store_remove(list_store, &list_iter);
    gtk_tree_selection_select_path(selection, tmppath);
  }
}

void restart_needed() {
  sprintf(gnome_dialog_msg, _("You need to restart Gjiten for some preferences to take effect.\n"));
}

void preferences_exit() {
  gtk_widget_destroy(dialog_preferences);
  dialog_preferences = NULL;
}

void pref_closed(GtkDialog *dialog, gint response, gpointer user_data) {
  //gchar *confpath;
  gchar *testbuffer;
  int shutdown_save;
  int i, bytesread, pos;
  gboolean valid;
  gchar *dicfile_name;
      
  if (response != GTK_RESPONSE_OK) {
    preferences_exit();
    return;
  }
  //gconf_client_add_dir(gconf_client, "/apps/gjiten", GCONF_CLIENT_PRELOAD_NONE, NULL);

  for (i = 0; i < KCFGNUM; i++) { 
    if (GTK_TOGGLE_BUTTON (checkb_prefs[i])->active) conf.kdiccfg[i] = TRUE;
    else conf.kdiccfg[i] = FALSE; 
  }
  if (strlen(gtk_entry_get_text(GTK_ENTRY(combo_entry_dictpath)))) {
    conf.dictpath = g_strdup(gtk_entry_get_text(GTK_ENTRY(combo_entry_dictpath)));
  }
  if (strlen(gtk_entry_get_text(GTK_ENTRY(combo_entry_kanjidic))))  {
    conf.kanjidicfile = g_strdup(gtk_entry_get_text(GTK_ENTRY(combo_entry_kanjidic)));
  }
  
  /*
  if ((GTK_TOGGLE_BUTTON (checkb_tooltips)->active)) conf.tooltips = TRUE;
  else conf.tooltips = FALSE;
  */
  if ((GTK_TOGGLE_BUTTON (checkb_menubar)->active)) conf.menubar = TRUE;
  else conf.menubar = FALSE;
  if ((GTK_TOGGLE_BUTTON (checkb_toolbar)->active)) conf.toolbar = TRUE;
  else conf.toolbar = FALSE;
  if ((GTK_TOGGLE_BUTTON (checkb_kanjipad)->active)) {
    conf.usekanjipad = TRUE;
    conf.kanjipad = g_strdup(gtk_entry_get_text(GTK_ENTRY(combo_entry_kanjipad)));
  }
  else conf.usekanjipad = FALSE;
  if ((GTK_TOGGLE_BUTTON (checkb_kanjidic)->active)) conf.usekanjidic = TRUE;
  else conf.usekanjidic = FALSE;
  conf.bigwords = GTK_TOGGLE_BUTTON(checkb_bigwords)->active;
  conf.bigkanji = GTK_TOGGLE_BUTTON(checkb_bigkanji)->active;
  conf.gdk_use_xft = GTK_TOGGLE_BUTTON(checkb_use_xft)->active;
  conf.force_ja_JP = GTK_TOGGLE_BUTTON(checkb_force_ja_JP)->active;
  conf.force_language_c = GTK_TOGGLE_BUTTON(checkb_force_language_c)->active;
  conf.normalfont = g_strdup(gtk_entry_get_text(GTK_ENTRY(entry_normalfont)));
  conf.bigfont = g_strdup(gtk_entry_get_text(GTK_ENTRY(entry_bigfont)));

  conf.numofdics = 0;
  valid = gtk_tree_model_get_iter_first (model, &list_iter);
  while (valid == TRUE) {
    gtk_tree_model_get(model, &list_iter, STRING_COLUMN, &dicfile_name, -1);
    conf.dicfiles[conf.numofdics] = dicfile_name;    
    conf.numofdics++;
    valid = gtk_tree_model_iter_next(model, &list_iter);
  }
 
  save_settings();
  load_settings();
  init_settings();

  //TEST if dictfiles are UTF-8
  //Test kanjidic
  if (conf.kanjidicfile_descriptor > 0) {
    if (conf.usekanjidic == TRUE) {
      //printf("kanjidic test.\n");
      testbuffer = (gchar *) g_malloc(3000);
      bytesread = read(conf.kanjidicfile_descriptor, testbuffer, 3000); // read a chunk into buffer
      pos = bytesread - 1;
      while (testbuffer[pos] != '\n') pos--;
      if (g_utf8_validate(testbuffer, pos, NULL) == FALSE) {
	sprintf(gnome_dialog_msg, 
		_("Dictionary file is non-UTF: %s\nPlease convert it to UTF-8. See the docs for more."), conf.kanjidicfile);
	conf.dicts_ok = FALSE;
	//printf("DICT ERR\n");
      }
      g_free(testbuffer);
    }
  }
  
  for (i = 0; i < conf.numofdics; i++) { 
    if (conf.dicfile_descriptor[i] > 0) {
      //printf("DICT TEST: %s\n", conf.dicfiles[i]);
      testbuffer = (gchar *) g_malloc(3000);
      bytesread = read(conf.dicfile_descriptor[i], testbuffer, 3000); // read a chunk into buffer
      pos = bytesread - 1;
      while (testbuffer[pos] != '\n') pos--;
      if (g_utf8_validate(testbuffer, pos, NULL) == FALSE) {
	sprintf(gnome_dialog_msg, 
		_("Dictionary file is non-UTF: %s\nPlease convert it to UTF-8. See the docs for more."), conf.dicfiles[i]);
	conf.dicts_ok = FALSE;
	//printf("DICT ERR\n");
	g_free(testbuffer);
	break;
      }
      g_free(testbuffer);
    }
  }

  preferences_exit();
  if (window_gjiten != NULL) {        /* redraw windows*/
    shutdown_save = shutdown_gjiten;
    shutdown_gjiten = FALSE;
    gtk_widget_destroy(window_gjiten);
    create_window_gjiten();
   shutdown_gjiten = shutdown_save;
  }
  if (window_kanjidic != NULL) {
    gtk_widget_destroy(window_kanjidic);
    create_window_kanjidic();
  }
}


void shade_pref_widgets() {
  gtk_widget_set_sensitive(combo_entry_kanjidic,GTK_TOGGLE_BUTTON(checkb_kanjidic)->active);
  gtk_widget_set_sensitive(combo_entry_kanjipad,GTK_TOGGLE_BUTTON(checkb_kanjipad)->active);
}


void set_strginfo() { /* Efficient way to do this ?? */
  strginfo[0] = _("Kanji");
  strginfo[1] = _("Radicals");
  strginfo[2] =	_("Stroke count");
  strginfo[3] =	_("Readings");
  strginfo[4] =	_("Romanized Korean reading"); 
  strginfo[5] =	_("Romanized Pinyin reading"); 
  strginfo[6] =	_("English meaning"); 
  strginfo[7] = _("Bushu radical number");
  strginfo[8] = _("Classical radical number");
  strginfo[9] = _("Frequency number");
  strginfo[10] = _("Jouyou grade level");
  strginfo[11] = _("De Roo code");
  strginfo[12] = _("Skip pattern code"); 
  strginfo[13] = _("Four Corner code");
  strginfo[14] = _("\"New Jp-En Char Dictionary\" index");
  strginfo[15] = _("Index in Nelson's \"Modern Reader's Char Dict\"");
  strginfo[16] = _("\"The New Nelson Char Dict\" index"); 
  strginfo[17] = _("Spahn&Hadamitzky index"); 
  strginfo[18] = _("Morohashi \"Daikanwajiten\" index"); 
  strginfo[19] = _("Morohashi \"Daikanwajiten\" volume.page number"); 
  strginfo[20] = _("\"A Guide To Remembering Jap. Chars\" index"); 
  strginfo[21] = _("Gakken Kanji Dictionary index"); 
  strginfo[22] = _("Index in \"Remembering The Kanji\""); 
  strginfo[23] = _("Index in \"Japanese Names\""); 
  strginfo[24] = _("Cross-reference code"); 
  strginfo[25] = _("Misclassification code");
  strginfo[26] = _("Unicode hex number of the kanji");
  strginfo[27] = _("ASCII JIS Code of Kanji");
}

void create_dialog_preferences() {
  GtkWidget *vbox_pref_main;
  GtkWidget *vbox_main;
  GtkWidget *notebook_pref;
  GtkWidget *fileentry_kanjidic;
  GtkWidget *fileentry_kanjipad;
  GtkWidget *fileentry_dictpath;
  GtkWidget *scrolledwin_list;
  GtkWidget *button_adddict;
  GtkWidget *button_downdict;
  GtkWidget *button_updict;
  GtkWidget *button_deldict;
  GtkWidget *label_gopt;
  GtkWidget *label_kanjiopt;
  GtkWidget *label_kanjidicpath;
  GtkWidget *label_kanjipad;
  GtkWidget *label_path;
  GtkWidget *table_goptions;
  GtkWidget *table_koptions;
  GtkWidget *dialog_action_area;
  GtkWidget *table;
  GtkWidget *label;
  GtkWidget *entry;
  GtkWidget *fontpicker;
  int trow = 0, tcol = 0;
  int i;

  if (dialog_preferences != NULL) return;
  /*gtk_window_set_policy (GTK_WINDOW (dialog_preferences), FALSE, FALSE, FALSE);*/
  // gtk_window_set_modal(GTK_WINDOW (dialog_preferences),TRUE);
  dialog_preferences =  gtk_dialog_new_with_buttons (_("Preferences"), GTK_WINDOW(window_gjiten),
						     GTK_DIALOG_DESTROY_WITH_PARENT,
						     GTK_STOCK_OK,
						     GTK_RESPONSE_OK,
						     GTK_STOCK_CANCEL,
						     GTK_RESPONSE_REJECT,
						     NULL);
  g_signal_connect (GTK_OBJECT(dialog_preferences), "delete_event",
		      GTK_SIGNAL_FUNC (preferences_exit), NULL);

  vbox_pref_main = GTK_DIALOG(dialog_preferences)->vbox;
  gtk_widget_show (vbox_pref_main);

  vbox_main = gtk_vbox_new (FALSE, 10);
  gtk_widget_show (vbox_main);
  gtk_box_pack_start (GTK_BOX (vbox_pref_main), vbox_main, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (vbox_main), 11);

  notebook_pref = gtk_notebook_new ();
  gtk_widget_show (notebook_pref);
  gtk_box_pack_start (GTK_BOX (vbox_pref_main), notebook_pref, TRUE, TRUE, 0);

  table_goptions = gtk_table_new (8, 4, FALSE);
  gtk_widget_show (table_goptions);
  gtk_container_add (GTK_CONTAINER (notebook_pref), table_goptions);

  fileentry_kanjidic = gnome_file_entry_new (NULL, NULL);
  gtk_widget_show(fileentry_kanjidic);
  gtk_table_attach(GTK_TABLE (table_goptions), fileentry_kanjidic, 2, 4, 1, 2,
		   (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
		   (GtkAttachOptions) (0), 0, 0);

  combo_entry_kanjidic = gnome_file_entry_gtk_entry (GNOME_FILE_ENTRY (fileentry_kanjidic));
  gtk_widget_show (combo_entry_kanjidic);
  if (conf.kanjidicfile != NULL)
    gtk_entry_set_text(GTK_ENTRY(combo_entry_kanjidic), conf.kanjidicfile);

  fileentry_kanjipad = gnome_file_entry_new (NULL, NULL);
  gtk_widget_show (fileentry_kanjipad);
  gtk_table_attach (GTK_TABLE (table_goptions), fileentry_kanjipad, 2, 4, 2, 3,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL), (GtkAttachOptions) (0), 0, 0);

  combo_entry_kanjipad = gnome_file_entry_gtk_entry(GNOME_FILE_ENTRY(fileentry_kanjipad));
  gtk_widget_show(combo_entry_kanjipad);
  if (conf.kanjipad != NULL)
    gtk_entry_set_text (GTK_ENTRY (combo_entry_kanjipad), conf.kanjipad);

  label_path = gtk_label_new (_("Path To Dictionary Files:"));
  gtk_widget_show (label_path);
  gtk_table_attach (GTK_TABLE (table_goptions), label_path, 1, 2, 3, 4,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_justify (GTK_LABEL (label_path), GTK_JUSTIFY_RIGHT);
  gtk_misc_set_alignment (GTK_MISC (label_path), 0.9, 0.5);

  /*
  checkb_tooltips = gtk_check_button_new_with_label (_("Tooltips"));
  gtk_widget_show (checkb_tooltips);
  gtk_table_attach (GTK_TABLE (table_goptions), checkb_tooltips, 0, 1, 0, 1,
                    (GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
  if (conf.tooltips) 
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (checkb_tooltips), TRUE);
  */

  checkb_toolbar = gtk_check_button_new_with_label (_("Toolbar Visible"));
  gtk_widget_show (checkb_toolbar);
  gtk_table_attach (GTK_TABLE (table_goptions), checkb_toolbar, 0, 1, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),  (GtkAttachOptions) (0), 0, 0);
  if (conf.toolbar) 
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (checkb_toolbar), TRUE);

  checkb_menubar = gtk_check_button_new_with_label (_("Menubar Visible"));
  gtk_widget_show (checkb_menubar);
  gtk_table_attach (GTK_TABLE (table_goptions), checkb_menubar, 1, 2, 0, 1,
                    (GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
  if (conf.menubar) 
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (checkb_menubar), TRUE);

  label_kanjidicpath = gtk_label_new (_("KanjiDic File:"));
  gtk_widget_show (label_kanjidicpath);
  gtk_table_attach (GTK_TABLE (table_goptions), label_kanjidicpath, 1, 2, 1, 2,
                    (GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_justify (GTK_LABEL (label_kanjidicpath), GTK_JUSTIFY_RIGHT);
  gtk_misc_set_alignment (GTK_MISC (label_kanjidicpath), 0.9, 0.5);

  label_kanjipad = gtk_label_new (_("KanjiPad Executable:"));
  gtk_widget_show (label_kanjipad);
  gtk_table_attach (GTK_TABLE (table_goptions), label_kanjipad, 1, 2, 2, 3,
                    (GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label_kanjipad), 0.9, 0.5);

  checkb_kanjipad = gtk_check_button_new_with_label (_("Use KanjiPad"));
  gtk_widget_show (checkb_kanjipad);
  gtk_table_attach (GTK_TABLE (table_goptions), checkb_kanjipad, 0, 1, 2, 3,
                    (GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);

  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (checkb_kanjipad), conf.usekanjipad);
  gtk_widget_set_sensitive(combo_entry_kanjipad, conf.usekanjipad);
  g_signal_connect (GTK_OBJECT (checkb_kanjipad), "toggled", 
		      GTK_SIGNAL_FUNC(shade_pref_widgets), NULL );

  checkb_kanjidic = gtk_check_button_new_with_label (_("Use KanjiDic"));
  gtk_widget_show (checkb_kanjidic);
  gtk_table_attach (GTK_TABLE (table_goptions), checkb_kanjidic, 0, 1, 1, 2,
                    (GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (checkb_kanjidic), conf.usekanjidic);
  gtk_widget_set_sensitive(combo_entry_kanjidic, conf.usekanjidic);

  g_signal_connect (GTK_OBJECT (checkb_kanjidic), "toggled", 
		      GTK_SIGNAL_FUNC(shade_pref_widgets), NULL );


  fileentry_dictpath = gnome_file_entry_new (NULL, NULL);
  gtk_widget_show (fileentry_dictpath);
  gtk_table_attach (GTK_TABLE (table_goptions), fileentry_dictpath, 2, 4, 3, 4,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),  (GtkAttachOptions) (0), 0, 0);

  combo_entry_dictpath = gnome_file_entry_gtk_entry (GNOME_FILE_ENTRY (fileentry_dictpath));
  gtk_widget_show (combo_entry_dictpath);
  if (conf.dictpath != NULL)
    gtk_entry_set_text (GTK_ENTRY (combo_entry_dictpath), conf.dictpath);

  scrolledwin_list = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwin_list);
  gtk_table_attach(GTK_TABLE (table_goptions), scrolledwin_list, 0, 3, 4, 8,
		   (GtkAttachOptions) (GTK_EXPAND | GTK_FILL), (GtkAttachOptions) (GTK_EXPAND | GTK_FILL), 0, 0);

  //Set up the dictfile list
  list_store = gtk_list_store_new (N_COLUMNS, G_TYPE_STRING, G_TYPE_INT);
  dict_list_view = gtk_tree_view_new_with_model(GTK_TREE_MODEL(list_store));
  for (i = 0; i < conf.numofdics; i++) {
    gtk_list_store_append(list_store, &list_iter);
    gtk_list_store_set(list_store, &list_iter, STRING_COLUMN, conf.dicfiles[i], -1);
  }
  model = gtk_tree_view_get_model(GTK_TREE_VIEW(dict_list_view));
  gtk_widget_show(dict_list_view);

  cell_renderer = gtk_cell_renderer_text_new();
  list_column = gtk_tree_view_column_new_with_attributes(_("Dictionary Files"), cell_renderer, "text", 
							 STRING_COLUMN, NULL);

  gtk_tree_view_append_column(GTK_TREE_VIEW(dict_list_view), list_column);

  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrolledwin_list), dict_list_view);

  button_adddict = gtk_button_new_from_stock(GTK_STOCK_ADD);
  gtk_widget_show (button_adddict);
  gtk_table_attach(GTK_TABLE (table_goptions), button_adddict, 3, 4, 4, 5,
		   (GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (GTK_FILL), 0, 0);
  g_signal_connect(GTK_OBJECT (button_adddict), "clicked", 
		     GTK_SIGNAL_FUNC(add_dict), NULL );
  

  button_deldict = gtk_button_new_from_stock(GTK_STOCK_DELETE);
  gtk_widget_show(button_deldict);
  gtk_table_attach (GTK_TABLE(table_goptions), button_deldict, 3, 4, 5, 6,
                    (GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (GTK_FILL), 0, 0);
  g_signal_connect(GTK_OBJECT(button_deldict), "clicked", 
		      GTK_SIGNAL_FUNC(del_dict), NULL );

  button_updict = gtk_button_new_from_stock(GTK_STOCK_GO_UP);
  gtk_widget_show(button_updict);
  gtk_table_attach(GTK_TABLE(table_goptions), button_updict, 3, 4, 6, 7,
                    (GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (GTK_FILL), 0, 0);
  g_signal_connect (GTK_OBJECT (button_updict), "clicked", 
		      GTK_SIGNAL_FUNC(up_dict), NULL );

  button_downdict = gtk_button_new_from_stock(GTK_STOCK_GO_DOWN);
  gtk_widget_show(button_downdict);
  gtk_table_attach(GTK_TABLE(table_goptions), button_downdict, 3, 4, 7, 8,
                    (GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (GTK_FILL), 0, 0);
  g_signal_connect(GTK_OBJECT(button_downdict), "clicked", 
		      GTK_SIGNAL_FUNC(down_dict), NULL );


  label_gopt = gtk_label_new (_("General Options"));
  gtk_widget_show (label_gopt);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook_pref), 
			      gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook_pref), 0), 
			      label_gopt);


  table_koptions = gtk_table_new (11, 3, FALSE);
  gtk_widget_show (table_koptions);
  gtk_container_add (GTK_CONTAINER (notebook_pref), table_koptions); 

  for (tcol = 0; (tcol * 14 + trow < KCFGNUM) || (tcol < 3); tcol++)
    for (trow = 0; (tcol * 14 + trow < KCFGNUM) && (trow < 14); trow++) {
      checkb_prefs[tcol * 14 + trow] = gtk_check_button_new_with_label (strginfo[tcol * 14 + trow]);
      gtk_widget_show (checkb_prefs[tcol * 14 + trow]);
      gtk_table_attach (GTK_TABLE (table_koptions), checkb_prefs[tcol * 14 + trow], tcol, tcol + 1, trow, trow + 1,
                    (GtkAttachOptions) (GTK_FILL),  (GtkAttachOptions) (0), 0, 0);
      if (conf.kdiccfg[tcol * 14 + trow] == TRUE) 
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (checkb_prefs[tcol * 14 + trow]), TRUE);
    }

  label_kanjiopt = gtk_label_new(_("KanjiInfo Options"));
  gtk_widget_show(label_kanjiopt);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook_pref), gtk_notebook_get_nth_page 
			      (GTK_NOTEBOOK (notebook_pref), 1), label_kanjiopt);


  table = gtk_table_new(3, 3, 0);
  gtk_widget_show(table);
  gtk_container_set_border_width(GTK_CONTAINER(table), 10);
  gtk_table_set_row_spacings(GTK_TABLE(table), 5);
  gtk_table_set_col_spacings(GTK_TABLE(table), 5);

  label = gtk_label_new(_("Normal font: "));
  gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_RIGHT);
  gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
  gtk_widget_show(label);
  gtk_table_attach(GTK_TABLE(table), label, 0, 1, 0, 1,
		   (GtkAttachOptions)(GTK_FILL),
		   (GtkAttachOptions)0, 0, 0);
  entry = gtk_entry_new();
  gtk_widget_show(entry);
  entry_normalfont = entry;
  gtk_table_attach(GTK_TABLE(table), entry, 1, 2, 0, 1,
		   (GtkAttachOptions)(GTK_EXPAND|GTK_FILL),
		   (GtkAttachOptions)0, 0, 0);
  fontpicker = gnome_font_picker_new();
  gtk_widget_show(fontpicker);
  if (conf.normalfont != NULL) {
    gtk_entry_set_text(GTK_ENTRY(entry), conf.normalfont);
    //gnome_font_picker_set_font_name(GNOME_FONT_PICKER(fontpicker), conf.normalfont);
  }

  gnome_font_picker_set_preview_text(GNOME_FONT_PICKER(fontpicker), fontteststrg);
  gnome_font_picker_set_mode(GNOME_FONT_PICKER(fontpicker), GNOME_FONT_PICKER_MODE_PIXMAP);
  g_signal_connect_swapped (GTK_OBJECT (fontpicker), "font-set", GTK_SIGNAL_FUNC(font_set), (gpointer)entry);
  gtk_table_attach(GTK_TABLE(table), fontpicker, 2, 3, 0, 1,
		   (GtkAttachOptions)(GTK_FILL),
		   (GtkAttachOptions)0, 0, 0);

  label = gtk_label_new(_("Big font: "));
  gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_RIGHT);
  gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
  gtk_widget_show(label);
  gtk_table_attach(GTK_TABLE(table), label, 0, 1, 1, 2,
		   (GtkAttachOptions)(GTK_FILL),
		   (GtkAttachOptions)0, 0, 0);
  entry = gtk_entry_new();
  gtk_widget_show(entry);
  entry_bigfont = entry;
  gtk_table_attach(GTK_TABLE(table), entry, 1, 2, 1, 2,
		   (GtkAttachOptions)(GTK_EXPAND|GTK_FILL),
		   (GtkAttachOptions)0, 0, 0);
  fontpicker = gnome_font_picker_new();
  gtk_widget_show(fontpicker);
  if (conf.bigfont != NULL) {
    gtk_entry_set_text(GTK_ENTRY(entry), conf.bigfont);
    //gnome_font_picker_set_font_name(GNOME_FONT_PICKER(fontpicker), conf.bigfont);
  }
  gnome_font_picker_set_preview_text(GNOME_FONT_PICKER(fontpicker), fontteststrg);
  gnome_font_picker_set_mode(GNOME_FONT_PICKER(fontpicker), GNOME_FONT_PICKER_MODE_PIXMAP);
  g_signal_connect_swapped (GTK_OBJECT (fontpicker), "font-set", GTK_SIGNAL_FUNC(font_set), (gpointer)entry);
  gtk_table_attach(GTK_TABLE(table), fontpicker, 2, 3, 1, 2,
		   (GtkAttachOptions)(GTK_FILL),
		   (GtkAttachOptions)0, 0, 0);
  
  checkb_bigwords = gtk_check_button_new_with_label (_("Use big font for japanese words in search results"));
  gtk_widget_show (checkb_bigwords);
  gtk_table_attach(GTK_TABLE(table), checkb_bigwords, 0, 3, 2, 3,
		   (GtkAttachOptions)(GTK_FILL),
		   (GtkAttachOptions)0, 0, 0);
  if (conf.bigwords) 
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (checkb_bigwords), TRUE);

  checkb_bigkanji = gtk_check_button_new_with_label(_("Use big font for kanji results in KanjiDic"));
  gtk_widget_show (checkb_bigkanji);
  gtk_table_attach(GTK_TABLE(table), checkb_bigkanji, 0, 3, 3, 4,
		   (GtkAttachOptions)(GTK_FILL),
		   (GtkAttachOptions)0, 0, 0);
  if (conf.bigkanji) 
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (checkb_bigkanji), TRUE);

  checkb_use_xft = gtk_check_button_new_with_label(_("Antialiasing (might not work with Japanese)"));
  gtk_widget_show(checkb_use_xft);
  gtk_table_attach(GTK_TABLE(table), checkb_use_xft, 0, 3, 4, 5,
		   (GtkAttachOptions)(GTK_FILL),
		   (GtkAttachOptions)0, 0, 0);
  if (conf.gdk_use_xft) {
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkb_use_xft), TRUE);
  }
  g_signal_connect(GTK_OBJECT(checkb_use_xft), "clicked", 
		   GTK_SIGNAL_FUNC(restart_needed), NULL );

  checkb_force_ja_JP = gtk_check_button_new_with_label(_("Force LC_ALL=ja_JP (for kinput2)"));
  gtk_widget_show(checkb_force_ja_JP);
  gtk_table_attach(GTK_TABLE(table), checkb_force_ja_JP, 0, 3, 5, 6,
		   (GtkAttachOptions)(GTK_FILL),
		   (GtkAttachOptions)0, 0, 0);
  if (conf.force_ja_JP) {
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkb_force_ja_JP), TRUE);
  }
  g_signal_connect(GTK_OBJECT(checkb_force_ja_JP), "clicked", 
		   GTK_SIGNAL_FUNC(restart_needed), NULL );

  checkb_force_language_c = gtk_check_button_new_with_label(_("Force LANGUAGE=C (for English messages and GUI text)"));
  gtk_widget_show(checkb_force_language_c);
  gtk_table_attach(GTK_TABLE(table), checkb_force_language_c, 0, 3, 6, 7,
		   (GtkAttachOptions)(GTK_FILL),
		   (GtkAttachOptions)0, 0, 0);
  if (conf.force_language_c) {
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkb_force_language_c), TRUE);
  }
  g_signal_connect(GTK_OBJECT(checkb_force_language_c), "clicked", 
		   GTK_SIGNAL_FUNC(restart_needed), NULL );

  gtk_container_add (GTK_CONTAINER (notebook_pref), table); 
  label = gtk_label_new (_("Fonts"));
  gtk_widget_show (label);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook_pref), gtk_notebook_get_nth_page 
			      (GTK_NOTEBOOK (notebook_pref), 2), label);
 
  dialog_action_area = GTK_DIALOG (dialog_preferences)->action_area;
  gtk_widget_show (dialog_action_area);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (dialog_action_area), GTK_BUTTONBOX_END);

  /*
  gtk_dialog_add_button(GTK_DIALOG(dialog_preferences), GTK_STOCK_OK);
  button_prefok = g_list_last (GTK_DIALOG (dialog_preferences)->buttons)->data;
  gtk_widget_show (button_prefok);
  GTK_WIDGET_SET_FLAGS (button_prefok, GTK_CAN_DEFAULT);
  g_signal_connect_swapped (GTK_OBJECT (button_prefok), "clicked", GTK_SIGNAL_FUNC(pref_ok), NULL);

  gtk_dialog_add_button(GTK_DIALOG(dialog_preferences), GTK_STOCK_CANCEL);
  button_prefcancel = g_list_last (GTK_DIALOG (dialog_preferences)->buttons)->data;
  gtk_widget_show (button_prefcancel);
  GTK_WIDGET_SET_FLAGS (button_prefcancel, GTK_CAN_DEFAULT);
  g_signal_connect_swapped (GTK_OBJECT (button_prefcancel), "clicked", 
			     GTK_SIGNAL_FUNC(preferences_exit), GTK_OBJECT(dialog_preferences) );
  */

  g_signal_connect_swapped (GTK_OBJECT (dialog_preferences), "response", GTK_SIGNAL_FUNC(pref_closed), NULL);

  gtk_widget_show(dialog_preferences);
  /*
  if (gtk_dialog_run(GTK_DIALOG(dialog_preferences)) == GTK_RESPONSE_OK) pref_ok();
  gtk_widget_destroy(dialog_preferences);
  dialog_preferences = NULL;
  */
}
